#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#  postinst.py
#
#  Copyright 2018 Stevan Radaković <stevan.radakovic@linaro.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import os
import psycopg2
import pwd
import random
import shutil
import glob
import subprocess  # nosec - controlled inputs.
import sys

from lava_server.settings.config_file import ConfigFile


LAVA_SYS_MOUNTDIR = "/var/lib/lava-server/default/"
LAVA_SYS_HOME = "/var/lib/lava-server/home/"
LAVA_LOGS = "/var/log/lava-server/"
INSTANCE_CONF = "/etc/lava-server/instance.conf"
INSTANCE_TEMPLATE_CONF = "/usr/share/lava-server/instance.conf.template"
SECRET_KEY = "/etc/lava-server/secret_key.conf"
DISPATCHER_CONFIG = "/etc/lava-server/dispatcher-config/"
LAVA_DB_SERVER = "localhost"
# pylint: disable=line-too-long,missing-docstring


def psql_run(cmd_list, failure_msg):
    uid = pwd.getpwnam("postgres")[2]
    os.seteuid(uid)
    ret = run(cmd_list, failure_msg)
    uid = pwd.getpwnam("root")[2]
    os.seteuid(uid)
    return ret


def run(cmd_list, failure_msg):
    print(" ".join(cmd_list))
    try:
        ret = subprocess.check_call(cmd_list)  # nosec - internal.
    except subprocess.CalledProcessError:
        print(failure_msg)
        # all failures are fatal during setup
        sys.exit(1)
    return ret


def db_setup(config, pg_admin_username, pg_admin_password):

    # check postgres is not just installed but actually ready.
    run(
        [
            "pg_isready",
            "-d",
            config.LAVA_DB_NAME,
            "-p",
            "%s" % config.LAVA_DB_PORT,
            "-q",
        ],
        "Failed to connect to postgres.",
    )

    conn = psycopg2.connect(
        "dbname='postgres' user='%s' host='%s' password='%s' connect_timeout=5"
        % (pg_admin_username, config.LAVA_DB_SERVER, pg_admin_password)
    )
    conn.autocommit = True
    # reuse the connection to manipulate DB.
    cursor = conn.cursor()

    try:
        cursor.execute(
            "CREATE ROLE \"%s\" NOSUPERUSER NOCREATEDB NOCREATEROLE INHERIT LOGIN ENCRYPTED PASSWORD '%s'"
            % (config.LAVA_DB_USER, config.LAVA_DB_PASSWORD)
        )
    except psycopg2.ProgrammingError as exc:
        print(exc)

    cursor.execute(
        "SELECT EXISTS(SELECT * FROM information_schema.tables WHERE table_name='%s')"
        % config.LAVA_DB_NAME
    )  # nosec - not accessible.
    db_existed_before = cursor.fetchone()[0]

    if not db_existed_before:
        try:
            cursor.execute(
                "CREATE DATABASE \"%s\" LC_CTYPE 'C.UTF-8' ENCODING 'UTF-8' OWNER \"%s\" TEMPLATE template0"
                % (config.LAVA_DB_NAME, config.LAVA_DB_USER)
            )
        except psycopg2.ProgrammingError as exc:
            print(exc)

    conn = psycopg2.connect(
        "dbname='%s' user='%s' host='%s' password='%s' connect_timeout=5"
        % (
            config.LAVA_DB_NAME,
            pg_admin_username,
            config.LAVA_DB_SERVER,
            pg_admin_password,
        )
    )
    conn.autocommit = True
    # reuse the connection to manipulate DB.
    cursor = conn.cursor()

    with contextlib.suppress(psycopg2.ProgrammingError):
        cursor.execute(
            "CREATE ROLE devel NOSUPERUSER CREATEDB NOCREATEROLE INHERIT LOGIN ENCRYPTED PASSWORD 'devel'"
        )

    with contextlib.suppress(psycopg2.ProgrammingError):
        cursor.execute("""CREATE DATABASE OWNER devel devel""")

    if db_existed_before:
        run(["lava-server", "manage", "drop_materialized_views"], "materialized views")

    run(
        ["lava-server", "manage", "migrate", "--noinput", "--fake-initial"], "migration"
    )

    print("Refreshing all materialized views: lava-server manage refresh_queries --all")
    run(["lava-server", "manage", "refresh_queries", "--all"], "refresh_queries")

    cursor.execute("""SELECT * FROM pg_authid WHERE rolname='lavaserver'""")
    lava_user = cursor.fetchone()

    if not lava_user:
        run(
            [
                "lava-server",
                "manage",
                "createsuperuser",
                "--noinput",
                "--username=%s" % config.LAVA_SYS_USER,
                "--email=%s@lava.invalid" % config.LAVA_SYS_USER,
            ],
            "create super user",
        )


def configure():

    if os.path.exists(INSTANCE_CONF) and os.path.isfile(INSTANCE_CONF):
        config_path = INSTANCE_CONF
    else:
        config_path = INSTANCE_TEMPLATE_CONF

    config = ConfigFile.load(config_path)

    config.LAVA_DB_SERVER = LAVA_DB_SERVER

    if not hasattr(config, "LAVA_SYS_USER"):
        config.LAVA_SYS_USER = "lavaserver"

    if not hasattr(config, "LAVA_INSTANCE") or config.LAVA_INSTANCE == "$LAVA_INSTANCE":
        config.LAVA_INSTANCE = "default"

    if not hasattr(config, "LAVA_DB_NAME") or config.LAVA_DB_NAME == "$LAVA_DB_NAME":
        config.LAVA_DB_NAME = "lavaserver"

    if not hasattr(config, "LAVA_DB_USER") or config.LAVA_DB_USER == "$LAVA_DB_USER":
        config.LAVA_DB_USER = "lavaserver"

    if not hasattr(config, "LAVA_DB_PORT") or config.LAVA_DB_PORT == "$LAVA_DB_PORT":
        config.LAVA_DB_PORT = 5432

    if (
        not hasattr(config, "LAVA_DB_PASSWORD")
        or config.LAVA_DB_PASSWORD == "$LAVA_DB_PASSWORD"
    ):
        config.LAVA_DB_PASSWORD = "%012x" % random.getrandbits(48)

    ConfigFile.serialize(INSTANCE_CONF, config.__dict__)

    os.makedirs("%s/media/job-output/" % LAVA_SYS_MOUNTDIR, exist_ok=True)

    run(
        [
            "adduser",
            "--quiet",
            "--system",
            "--group",
            "--home=%s" % LAVA_SYS_HOME,
            config.LAVA_SYS_USER,
            "--shell=/bin/sh",
        ],
        "adduser",
    )

    shutil.chown(LAVA_SYS_MOUNTDIR, config.LAVA_SYS_USER, config.LAVA_SYS_USER)
    shutil.chown(
        "%s/media/" % LAVA_SYS_MOUNTDIR, config.LAVA_SYS_USER, config.LAVA_SYS_USER
    )
    shutil.chown(
        "%s/media/job-output/" % LAVA_SYS_MOUNTDIR,
        config.LAVA_SYS_USER,
        config.LAVA_SYS_USER,
    )

    # fixup bug from date based subdirectories - allowed to be missing.
    try:
        shutil.chown(
            "%s/media/job-output/2017" % LAVA_SYS_MOUNTDIR,
            config.LAVA_SYS_USER,
            config.LAVA_SYS_USER,
        )
    except FileNotFoundError:
        print("legacy directory is missing, skip..")

    # support changes in xml-rpc API for 2017.6
    shutil.chown(
        "/etc/lava-server/dispatcher.d/", config.LAVA_SYS_USER, config.LAVA_SYS_USER
    )

    shutil.chown(DISPATCHER_CONFIG, config.LAVA_SYS_USER, config.LAVA_SYS_USER)

    shutil.chown(
        "%s/devices/" % DISPATCHER_CONFIG, config.LAVA_SYS_USER, config.LAVA_SYS_USER
    )

    shutil.chown(
        "%s/device-types/" % DISPATCHER_CONFIG,
        config.LAVA_SYS_USER,
        config.LAVA_SYS_USER,
    )

    # user may not have been removed but the directory has, after purge.
    if not os.path.isdir(LAVA_SYS_HOME):
        os.mkdir(LAVA_SYS_HOME)
        shutil.chown(LAVA_SYS_HOME, config.LAVA_SYS_USER, config.LAVA_SYS_USER)

    # Fix permissions of /etc/lava-server/instance.conf
    shutil.chown(INSTANCE_CONF, config.LAVA_SYS_USER, config.LAVA_SYS_USER)
    os.chmod(INSTANCE_CONF, 0o640)

    os.makedirs(os.path.dirname(LAVA_LOGS), exist_ok=True)

    # Allow lavaserver to write to all the log files
    # setgid on LAVA_LOGS directory
    os.chmod(LAVA_LOGS, 0o2775)  # nosec - group permissive.

    # Allow users in the adm group to read all logs
    with open("%s/django.log" % LAVA_LOGS, "w+") as logfile:
        logfile.write("")
    shutil.chown(LAVA_LOGS, user=config.LAVA_SYS_USER, group="adm")
    for file in glob.glob("%s/*" % LAVA_LOGS):
        if "lava-scheduler.log" in file:
            # skip changes to old logs.
            continue
        shutil.chown(file, user=config.LAVA_SYS_USER, group="adm")
        # allow users in the adm group to run lava-server commands
        os.chmod(file, 0o0664)

    # tidy up old logrotate config to allow logrotate cron to complete.
    if os.path.exists("/etc/logrotate.d/lava-scheduler-log"):
        os.unlink("/etc/logrotate.d/lava-scheduler-log")

    # Allow lava user to write the secret key
    with open(SECRET_KEY, "w+") as key:
        key.write("")
    shutil.chown(SECRET_KEY, config.LAVA_SYS_USER, config.LAVA_SYS_USER)
    os.chmod(SECRET_KEY, 0o640)

    # Allow lavaserver to write dispatcher-config files for XMLRPC API
    os.makedirs("%s/devices/" % DISPATCHER_CONFIG, exist_ok=True)
    shutil.chown(
        "%s/devices/" % DISPATCHER_CONFIG, config.LAVA_SYS_USER, config.LAVA_SYS_USER
    )
    for filename in glob.glob("%s/devices/*" % DISPATCHER_CONFIG):
        shutil.chown(filename, config.LAVA_SYS_USER, config.LAVA_SYS_USER)
    shutil.chown(
        "%s/health-checks/" % DISPATCHER_CONFIG,
        config.LAVA_SYS_USER,
        config.LAVA_SYS_USER,
    )
    for filename in glob.glob("%s/health-checks/*" % DISPATCHER_CONFIG):
        shutil.chown(filename, config.LAVA_SYS_USER, config.LAVA_SYS_USER)
    shutil.chown(
        "%s/device-types/" % DISPATCHER_CONFIG,
        config.LAVA_SYS_USER,
        config.LAVA_SYS_USER,
    )
    for filename in glob.glob("%s/device-types/*" % DISPATCHER_CONFIG):
        shutil.chown(filename, config.LAVA_SYS_USER, config.LAVA_SYS_USER)

    # Create temporary database role for db operations.
    pg_admin_username = "user_%012x" % random.getrandbits(48)
    pg_admin_password = "%012x" % random.getrandbits(48)

    result = psql_run(
        [
            "psql",
            "-c",
            "CREATE ROLE %s PASSWORD '%s' SUPERUSER CREATEDB CREATEROLE INHERIT LOGIN;"
            % (pg_admin_username, pg_admin_password),
        ],
        "Failed to create temporary superuser role",
    )

    if result != 0:
        print("Failed to create postgres superuser.")
        return

    try:
        db_setup(config, pg_admin_username, pg_admin_password)
    finally:
        # Removing temprorary user from postgres.
        result = psql_run(
            ["psql", "-c", "DROP ROLE %s ;" % pg_admin_username],
            "Failed to drop temporary superuser role.",
        )
        if result != 0:
            print(
                "Temporary user %s was not properly removed from postgres. Please do so manually."
                % pg_admin_username
            )


def main():
    configure()


if __name__ == "__main__":
    sys.exit(main())



# Copyright (C) 2016 Linaro Limited
#
# Author: Neil Williams <neil.williams@linaro.org>
#
# This file is part of LAVA Dispatcher.
#
# LAVA Dispatcher is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# LAVA Dispatcher is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along
# with this program; if not, see <http://www.gnu.org/licenses>.


import os
from lava_common.constants import INSTALLER_QUIET_MSG
from lava_common.exceptions import ConfigurationError, JobError
from lava_dispatcher.action import Action, Pipeline
from lava_dispatcher.logical import Boot
from lava_dispatcher.actions.boot import BootAction
from lava_dispatcher.utils.shell import which
from lava_dispatcher.utils.strings import substitute
from lava_dispatcher.actions.boot.environment import ExportDeviceEnvironment
from lava_dispatcher.shell import ExpectShellSession, ShellCommand, ShellSession
from lava_dispatcher.actions.boot import AutoLoginAction


class BootIsoInstaller(Boot):

    compatibility = 3

    def __init__(self, parent, parameters):
        super().__init__(parent)
        self.action = BootIsoInstallerAction()
        self.action.section = self.action_type
        self.action.job = self.job
        parent.add_action(self.action, parameters)

    @classmethod
    def accepts(cls, device, parameters):
        if "media" in parameters and parameters["media"] == "img":
            if "method" in parameters and parameters["method"] == "qemu-iso":
                return True, "accepted"
        return False, '"media" was not in parameters or "media" was not "img"'


class BootIsoInstallerAction(BootAction):

    name = "boot-installer-iso"
    description = "boot installer with preseed"
    summary = "boot installer iso image"

    def populate(self, parameters):
        self.internal_pipeline = Pipeline(
            parent=self, job=self.job, parameters=parameters
        )
        self.internal_pipeline.add_action(IsoCommandLine())
        self.internal_pipeline.add_action(MonitorInstallerSession())
        self.internal_pipeline.add_action(IsoRebootAction())
        # Add AutoLoginAction unconditionally as this action does nothing if
        # the configuration does not contain 'auto_login'
        self.internal_pipeline.add_action(AutoLoginAction())
        self.internal_pipeline.add_action(ExpectShellSession())
        self.internal_pipeline.add_action(ExportDeviceEnvironment())


class IsoCommandLine(Action):  # pylint: disable=too-many-instance-attributes

    """
    qemu-system-x86_64 -nographic -enable-kvm -cpu host -net nic,model=virtio,macaddr=52:54:00:12:34:59 -net user -m 2048 \
    -drive format=raw,file=hd_img.img -drive file=${NAME},index=2,media=cdrom,readonly \
    -boot c -no-reboot -kernel vmlinuz -initrd initrd.gz \
    -append "\"${BASE} ${LOCALE} ${CONSOLE} ${KEYMAPS} ${NETCFG} preseed/url=${PRESEED_URL} --- ${CONSOLE}\"" \
    """

    name = "execute-installer-command"
    description = "add dynamic data values to command line and execute"
    summary = "include downloaded locations and call qemu"

    def run(self, connection, max_end_time):
        # substitutions
        substitutions = {
            "{emptyimage}": self.get_namespace_data(
                action="prepare-empty-image", label="prepare-empty-image", key="output"
            )
        }
        sub_command = self.get_namespace_data(
            action="prepare-qemu-commands",
            label="prepare-qemu-commands",
            key="sub_command",
        )
        sub_command = substitute(sub_command, substitutions)
        command_line = " ".join(sub_command)

        commands = []
        # get the download args in run()
        image_arg = self.get_namespace_data(
            action="download-action", label="iso", key="image_arg"
        )
        action_arg = self.get_namespace_data(
            action="download-action", label="iso", key="file"
        )
        substitutions["{%s}" % "iso"] = action_arg
        commands.append(image_arg)
        command_line += " ".join(substitute(commands, substitutions))

        preseed_file = self.get_namespace_data(
            action="download-action", label="file", key="preseed"
        )
        if not preseed_file:
            raise JobError("Unable to identify downloaded preseed filename.")
        substitutions = {"{preseed}": preseed_file}
        append_args = self.get_namespace_data(
            action="prepare-qemu-commands", label="prepare-qemu-commands", key="append"
        )
        append_args = substitute([append_args], substitutions)
        command_line += " ".join(append_args)

        self.logger.info(command_line)
        shell = ShellCommand(command_line, self.timeout, logger=self.logger)
        if shell.exitstatus:
            raise JobError(
                "%s command exited %d: %s"
                % (sub_command[0], shell.exitstatus, shell.readlines())
            )
        self.logger.debug("started a shell command")

        shell_connection = ShellSession(self.job, shell)
        shell_connection.prompt_str = self.get_namespace_data(
            action="prepare-qemu-commands", label="prepare-qemu-commands", key="prompts"
        )
        shell_connection = super().run(shell_connection, max_end_time)
        return shell_connection


class MonitorInstallerSession(Action):
    """
    Waits for a shell connection to the device for the current job.
    The shell connection can be over any particular connection,
    all that is needed is a prompt.
    """

    compatibility = 3
    name = "monitor-installer-connection"
    description = "Monitor installer operation"
    summary = "Watch for error strings or end of install"

    def __init__(self):
        super().__init__()
        self.force_prompt = True

    def validate(self):
        super().validate()
        if "prompts" not in self.parameters:
            self.errors = "Unable to identify test image prompts from parameters."

    def run(self, connection, max_end_time):
        self.logger.debug(
            "%s: Waiting for prompt %s", self.name, " ".join(connection.prompt_str)
        )
        self.wait(connection, max_end_time)
        return connection


class IsoRebootAction(Action):

    name = "reboot-into-installed"
    description = "reboot and login to the new system"
    summary = "reboot into installed image"

    def __init__(self):
        super().__init__()
        self.sub_command = None

    def validate(self):
        super().validate()
        if "prompts" not in self.parameters:
            self.errors = "Unable to identify boot prompts from job definition."
        try:
            boot = self.job.device["actions"]["boot"]["methods"]["qemu"]
            qemu_binary = which(boot["parameters"]["command"])
            self.sub_command = [qemu_binary]
            self.sub_command.extend(boot["parameters"].get("options", []))
        except AttributeError as exc:
            raise ConfigurationError(exc)
        except (KeyError, TypeError):
            self.errors = "Invalid parameters for %s" % self.name

    def run(self, connection, max_end_time):
        """
        qemu needs help to reboot after running the debian installer
        and typically the boot is quiet, so there is almost nothing to log.
        """
        base_image = self.get_namespace_data(
            action="prepare-empty-image", label="prepare-empty-image", key="output"
        )
        self.sub_command.append("-drive format=raw,file=%s" % base_image)
        guest = self.get_namespace_data(
            action="apply-overlay-guest", label="guest", key="filename"
        )
        if guest:
            self.logger.info("Extending command line for qcow2 test overlay")
            self.sub_command.append(
                "-drive format=qcow2,file=%s,media=disk" % (os.path.realpath(guest))
            )
            # push the mount operation to the test shell pre-command to be run
            # before the test shell tries to execute.
            shell_precommand_list = []
            mountpoint = self.get_namespace_data(
                action="test", label="results", key="lava_test_results_dir"
            )
            shell_precommand_list.append("mkdir %s" % mountpoint)
            shell_precommand_list.append("mount -L LAVA %s" % mountpoint)
            self.set_namespace_data(
                action="test",
                label="lava-test-shell",
                key="pre-command-list",
                value=shell_precommand_list,
            )

        self.logger.info("Boot command: %s", " ".join(self.sub_command))
        shell = ShellCommand(
            " ".join(self.sub_command), self.timeout, logger=self.logger
        )
        if shell.exitstatus:
            raise JobError(
                "%s command exited %d: %s"
                % (self.sub_command, shell.exitstatus, shell.readlines())
            )
        self.logger.debug("started a shell command")

        shell_connection = ShellSession(self.job, shell)
        shell_connection = super().run(shell_connection, max_end_time)
        shell_connection.prompt_str = [INSTALLER_QUIET_MSG]
        self.wait(shell_connection)
        self.set_namespace_data(
            action="shared", label="shared", key="connection", value=shell_connection
        )
        return shell_connection



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('dashboard_app', '0007_imagereportchart_chart_visibility'),
    ]

    operations = [
        migrations.AddField(
            model_name='imagechartfilter',
            name='is_all_tests_included',
            field=models.BooleanField(default=False, verbose_name='Include all tests from this filter'),
            preserve_default=True,
        ),
    ]



# Copyright (C) 2018 Linaro Limited
#
# Author: Neil Williams <neil.williams@linaro.org>
#
# This file is part of LAVA.
#
# LAVA is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# LAVA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along
# with this program; if not, see <http://www.gnu.org/licenses>.

import datetime
import time
import signal
from contextlib import contextmanager
from lava_common.constants import ACTION_TIMEOUT
from lava_common.exceptions import JobError, ConfigurationError


class Timeout:
    """
    The Timeout class is a declarative base which any actions can use. If an Action has
    a timeout, that timeout name and the duration will be output as part of the action
    description and the timeout is then exposed as a modifiable value via the device_type,
    device or even job inputs. All timeouts are subject to a hardcoded maximum duration which
    cannot be exceeded by device_type, device or job input, only by the Action initialising
    the timeout.
    If a connection is set, this timeout is used per pexpect operation on that connection.
    If a connection is not set, this timeout applies for the entire run function of the action.
    """

    def __init__(self, name, duration=ACTION_TIMEOUT, exception=JobError):
        self.name = name
        self.start = 0
        self.elapsed_time = -1
        self.duration = duration  # Actions can set timeouts higher than the clamp.
        self.exception = exception

    @classmethod
    def default_duration(cls):
        return ACTION_TIMEOUT

    @classmethod
    def parse(cls, data):
        """
        Parsed timeouts can be set in device configuration or device_type configuration
        and can therefore exceed the clamp.
        """
        if not isinstance(data, dict):
            raise ConfigurationError("Invalid timeout data")
        duration = datetime.timedelta(
            days=data.get("days", 0),
            hours=data.get("hours", 0),
            minutes=data.get("minutes", 0),
            seconds=data.get("seconds", 0),
        )
        if not duration:
            return Timeout.default_duration()
        return int(duration.total_seconds())

    def can_skip(self, parameters):  # pylint: disable=no-self-use
        return parameters.get("timeout", {}).get("skip", False)

    def _timed_out(self, signum, frame):  # pylint: disable=unused-argument
        duration = int(time.time() - self.start)
        raise self.exception("%s timed out after %s seconds" % (self.name, duration))

    @contextmanager
    def __call__(self, parent, action_max_end_time):
        self.start = time.time()

        max_end_time = self.start + self.duration
        if action_max_end_time is not None:
            # action_max_end_time is None when called by the job class directly
            max_end_time = min(action_max_end_time, max_end_time)

        duration = round(max_end_time - self.start)
        if duration <= 0:
            # If duration is lower than 0, then the timeout should be raised now.
            # Calling signal.alarm in this case will only deactivate the alarm
            # (by passing 0 or the unsigned value).
            # Deactivate any previous alarm and set elapse_time prior to raise
            signal.alarm(0)
            self.elapsed_time = 0
            self._timed_out(None, None)

        signal.signal(signal.SIGALRM, self._timed_out)
        signal.alarm(duration)

        try:
            yield max_end_time

            # Restore the parent handler and timeout
            # This will be None when called by Job class
            if parent is None:
                signal.alarm(0)
            else:
                signal.signal(
                    signal.SIGALRM, parent.timeout._timed_out
                )  # pylint: disable=protected-access
                duration = round(action_max_end_time - time.time())
                if duration <= 0:
                    signal.alarm(0)
                    parent.timeout._timed_out(
                        None, None
                    )  # pylint: disable=protected-access
                signal.alarm(duration)
        except Exception:
            # clear the timeout alarm, the action has returned an error
            signal.alarm(0)
            raise
        finally:
            self.elapsed_time = time.time() - self.start



# Copyright (C) 2016 Linaro Limited
#
# Author: Matthew Hart <matthew.hart@linaro.org>
#
# This file is part of LAVA Dispatcher.
#
# LAVA Dispatcher is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# LAVA Dispatcher is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along
# with this program; if not, see <http://www.gnu.org/licenses>.

import re


def add_to_kickstart(preseedfile, extra_command):
    with open(preseedfile, "a") as pf:
        pf.write("\n")
        pf.write("%post\n")
        pf.write("exec < /dev/console > /dev/console\n")
        pf.write(extra_command + "\n")
        pf.write("%end\n")
    pf.close()


def add_late_command(preseedfile, extra_command):
    added = False
    with open(preseedfile, "r") as pf:
        lines = pf.readlines()
        pf.close()
    endstring = "\\\n"
    while endsin(lines, endstring):
        for linenum, data in enumerate(lines):
            if endsin(data, endstring):
                lines[linenum] = (
                    lines[linenum].replace(endstring, "") + lines[linenum + 1]
                )
                del lines[linenum + 1]
    for linenum, data in enumerate(lines):
        if re.match("d-i preseed/late_command string(.*)", data):
            # late_command already exists, append to it
            append_line = "; " + extra_command + "\n"
            lines[linenum] = lines[linenum].rstrip(" ;\n") + append_line
            added = True
    if not added:
        append_line = extra_command + "\n"
        lines.append("d-i preseed/late_command string " + append_line)

    with open(preseedfile, "w") as pf:
        for line in lines:
            pf.write(line)
        pf.close()


def endsin(lines, endstring):
    match = False
    if type(lines) is list:
        for line in lines:
            if line.endswith(endstring):
                match = True
    elif type(lines) is str:
        if lines.endswith(endstring):
            match = True
    return match



#! /usr/bin/python3

"""
This script is to check the combination of the Jinja2 device-type templates
and the instance-specific device dictionary configuration, to ensure that the
device configuration is valid YAML syntax for each device.

"""

# Copyright 2017 Linaro Limited
# Author: Neil Williams <neil.williams@linaro.org>
#         Remi Duraffort <remi.duraffort@linaro.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# pylint gets confused: commands have no shebang, but the file is not a module.
# pylint: disable=invalid-name

import argparse
import jinja2
import jinja2.exceptions
import glob
import os
import sys
import yaml


def main():
    parser = argparse.ArgumentParser(description="Check device templates")
    parser.add_argument(
        "--device-types",
        required=True,
        type=str,
        help="Path to the directory containing the device-type jinja2 templates.",
    )
    parser.add_argument(
        "--devices",
        required=True,
        type=str,
        help="Path to directory containing the device dictionary files.",
    )
    args = parser.parse_args()

    if not os.path.isdir(args.devices):
        sys.stderr.write("--devices argument must be a directory\n")
        return 1
    if not os.path.isdir(args.device_types):
        sys.stderr.write("--device-types argument must be a directory\n")
        return 1

    errors = False
    devices = sorted(glob.glob("%s/*.jinja2" % args.devices))

    print("Devices:")
    env = jinja2.Environment(  # nosec - YAML, not HTML, no XSS scope.
        loader=jinja2.FileSystemLoader([args.devices, args.device_types]),
        autoescape=False,
    )

    for device in devices:
        device_name = os.path.splitext(os.path.basename(device))[0]

        try:
            template = env.get_template("%s.jinja2" % device_name)
            device_template = template.render()
        except jinja2.exceptions.TemplateNotFound as exc:
            print('* %s (ERROR): "%s" not found' % (device_name, exc))
            errors = True
        except jinja2.exceptions.TemplateRuntimeError as exc:
            print('* %s (ERROR): rendering error "%s"' % (device_name, exc))
            errors = True
        except jinja2.exceptions.TemplateSyntaxError as exc:
            print(
                '* %s (ERROR): invalid syntax "%s" in "%s"'
                % (device_name, exc, exc.filename)
            )
            errors = True
        else:
            print("* %s" % device_name)

    return errors


if __name__ == "__main__":
    sys.exit(main())



# Copyright (C) 2015 Linaro Limited
#
# Author: Senthil Kumaran S <senthil.kumaran@linaro.org>
#
# This file is part of LAVA Dispatcher.
#
# LAVA Dispatcher is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# LAVA Dispatcher is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along
# with this program; if not, see <http://www.gnu.org/licenses>.

import time
from lava_common.exceptions import JobError
from lava_dispatcher.action import Pipeline, Action
from lava_dispatcher.logical import Boot
from lava_dispatcher.actions.boot import BootAction
from lava_dispatcher.actions.boot.environment import ExportDeviceEnvironment
from lava_dispatcher.connections.lxc import ConnectLxc
from lava_dispatcher.shell import ExpectShellSession
from lava_dispatcher.utils.shell import which
from lava_dispatcher.utils.udev import get_udev_devices
from lava_dispatcher.utils.udev import allow_fs_label


class BootLxc(Boot):
    """
    Attaches to the lxc container.
    """

    compatibility = 1

    def __init__(self, parent, parameters):
        super().__init__(parent)
        self.action = BootLxcAction()
        self.action.section = self.action_type
        self.action.job = self.job
        parent.add_action(self.action, parameters)

    @classmethod
    def accepts(cls, device, parameters):
        if "method" in parameters:
            if parameters["method"] == "lxc":
                return True, "accepted"
        return False, '"method" was not in parameters or "method" was not "lxc"'


class BootLxcAction(BootAction):
    """
    Provide for auto_login parameters in this boot stanza and re-establish the
    connection after boot.
    """

    name = "lxc-boot"
    description = "lxc boot into the system"
    summary = "lxc boot"

    def populate(self, parameters):
        self.internal_pipeline = Pipeline(
            parent=self, job=self.job, parameters=parameters
        )
        self.internal_pipeline.add_action(LxcStartAction())
        self.internal_pipeline.add_action(LxcAddStaticDevices())
        self.internal_pipeline.add_action(ConnectLxc())
        # Skip AutoLoginAction unconditionally as this action tries to parse kernel message
        # self.internal_pipeline.add_action(AutoLoginAction())
        self.internal_pipeline.add_action(ExpectShellSession())
        self.internal_pipeline.add_action(ExportDeviceEnvironment())


class LxcAddStaticDevices(Action):
    """
    Identifies permanently powered devices which are relevant
    to this LXC and adds the devices to the LXC after startup.
    e.g. Devices providing a tty are often powered from the
    worker.
    """

    name = "lxc-add-static"
    description = "Add devices which are permanently powered by the worker to the LXC"
    summary = "Add static devices to the LXC"

    def get_usb_devices(self):
        """
        Takes static_info from the device, and identifies which
        devices are USB devices. Only passes the USB devices back.
        """
        usb_devices = []
        for device in self.job.device.get("static_info", []):
            if "board_id" in device or "fs_label" in device:
                # This is a USB device
                usb_devices.append(device)
        return usb_devices

    def validate(self):
        super().validate()
        # If there are no USB devices under static_info then this action should be idempotent.

        # If we are allowed to use a filesystem label, we don't require a board_id
        # By default, we do require a board_id (serial)
        requires_board_id = not allow_fs_label(self.job.device)
        try:
            for usb_device in self.get_usb_devices():
                if (
                    usb_device.get("board_id", "") in ["", "0000000000"]
                    and requires_board_id
                ):
                    self.errors = "[LXC_STATIC] board_id unset"
                if usb_device.get("usb_vendor_id", "") == "0000":
                    self.errors = "[LXC_STATIC] usb_vendor_id unset"
                if usb_device.get("usb_product_id", "") == "0000":
                    self.errors = "[LXC_STATIC] usb_product_id unset"
        except TypeError:
            self.errors = "Invalid parameters for %s" % self.name

    def run(self, connection, max_end_time):
        connection = super().run(connection, max_end_time)
        lxc_name = self.get_namespace_data(
            action="lxc-create-action", label="lxc", key="name"
        )
        # If there are no USB devices under static_info then this action should be idempotent.
        if not self.get_usb_devices():
            return connection
        device_list = get_udev_devices(
            job=self.job, logger=self.logger, device_info=self.get_usb_devices()
        )
        for link in device_list:
            lxc_cmd = ["lxc-device", "-n", lxc_name, "add", link]
            cmd_out = self.run_command(lxc_cmd, allow_silent=True)
            if not isinstance(cmd_out, bool) and cmd_out:
                self.logger.debug(cmd_out)
        return connection


class LxcStartAction(Action):
    """
    This action calls lxc-start to get into the system.
    """

    name = "boot-lxc"
    description = "boot into lxc container"
    summary = "attempt to boot"

    def __init__(self):
        super().__init__()
        self.sleep = 10

    def validate(self):
        super().validate()
        which("lxc-start")

    def run(self, connection, max_end_time):
        connection = super().run(connection, max_end_time)
        lxc_name = self.get_namespace_data(
            action="lxc-create-action", label="lxc", key="name"
        )
        lxc_cmd = ["lxc-start", "-n", lxc_name, "-d"]
        command_output = self.run_command(lxc_cmd)
        if command_output and command_output != "":
            raise JobError(
                "Unable to start lxc container: %s" % command_output
            )  # FIXME: JobError needs a unit test
        lxc_cmd = ["lxc-info", "-sH", "-n", lxc_name]
        self.logger.debug("Wait until '%s' state becomes RUNNING", lxc_name)
        while True:
            command_output = self.run_command(lxc_cmd, allow_fail=True)
            if command_output and "RUNNING" in command_output.strip():
                break
            time.sleep(self.sleep)  # poll every 10 seconds.
        self.logger.info("'%s' state is RUNNING", lxc_name)
        # Check if LXC got an IP address so that we are sure, networking is
        # enabled and the LXC can update or install software.
        lxc_cmd = ["lxc-info", "-iH", "-n", lxc_name]
        self.logger.debug("Wait until '%s' gets an IP address", lxc_name)
        while True:
            command_output = self.run_command(lxc_cmd, allow_fail=True)
            if command_output:
                break
            time.sleep(self.sleep)  # poll every 10 seconds.
        self.logger.info("'%s' IP address is: '%s'", lxc_name, command_output.strip())
        return connection


class LxcStopAction(Action):
    """
    This action calls lxc-stop to stop the container.
    """

    name = "lxc-stop"
    description = "stop the lxc container"
    summary = "stop lxc"

    def validate(self):
        super().validate()
        which("lxc-stop")

    def run(self, connection, max_end_time):
        connection = super().run(connection, max_end_time)
        lxc_name = self.get_namespace_data(
            action="lxc-create-action", label="lxc", key="name"
        )
        lxc_cmd = ["lxc-stop", "-k", "-n", lxc_name]
        command_output = self.run_command(lxc_cmd)
        if command_output and command_output != "":
            raise JobError(
                "Unable to stop lxc container: %s" % command_output
            )  # FIXME: JobError needs a unit test
        return connection



# Copyright (C) 2018 Linaro Limited
#
# Author: Remi Duraffort <remi.duraffort@linaro.org>
#
# This file is part of LAVA Dispatcher.
#
# LAVA Dispatcher is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# LAVA Dispatcher is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along
# with this program; if not, see <http://www.gnu.org/licenses>.

import unittest

from lava_common.exceptions import InfrastructureError, JobError
from lava_dispatcher.tests.test_basic import Factory, StdoutTestCase
from lava_dispatcher.utils.shell import which


def check_docker():
    try:
        which("docker")
        return False
    except InfrastructureError:
        return True


def check_gdb_multipart():
    try:
        which("gdb-multiarch")
        return False
    except InfrastructureError:
        return True


class GDBFactory(Factory):
    def create_cc3230SF_job(self, filename):
        return self.create_job("cc3220SF-01.jinja2", filename)

    def create_cc3230SF_docker_job(self, filename):
        return self.create_job("cc3220SF-docker-01.jinja2", filename)


class TestGDB(StdoutTestCase):
    @unittest.skipIf(check_docker(), "docker not available")
    @unittest.skipIf(check_gdb_multipart(), "gdb-multiarch not available")
    def test_validate_errors(self):
        factory = GDBFactory()
        job = factory.create_cc3230SF_job("sample_jobs/cc3220SF-docker.yaml")
        with self.assertRaises(JobError):
            job.validate()
        self.assertEqual(
            job.pipeline.errors,
            ["Requesting a docker container while docker is not used for this device"],
        )

    @unittest.skipIf(check_docker(), "docker not available")
    @unittest.skipIf(check_gdb_multipart(), "gdb-multiarch not available")
    def test_without_docker(self):
        factory = GDBFactory()
        job = factory.create_cc3230SF_job("sample_jobs/cc3220SF.yaml")
        job.validate()
        description_ref = self.pipeline_reference("cc3220SF.yaml", job=job)
        self.assertEqual(description_ref, job.pipeline.describe(False))

        # Check BootGDBRetry action
        action = job.pipeline.actions[1].internal_pipeline.actions[0]
        self.assertEqual(action.name, "boot-gdb-retry")
        self.assertEqual(action.gdb, "gdb-multiarch")
        self.assertEqual(action.arguments, ["{ZEPHYR}"])
        self.assertEqual(len(action.commands), 4)
        self.assertEqual(
            action.commands[0],
            'target remote | openocd -c "gdb_port pipe" -f {OPENOCD_SCRIPT}',
        )
        self.assertEqual(action.commands[1], "monitor reset halt")
        self.assertEqual(action.commands[2], "load")
        self.assertEqual(action.commands[3], "set remotetimeout 10000")
        self.assertEqual(action.container, None)
        self.assertEqual(action.devices, [])

    @unittest.skipIf(check_docker(), "docker not available")
    @unittest.skipIf(check_gdb_multipart(), "gdb-multiarch not available")
    def test_with_docker(self):
        factory = GDBFactory()

        # Specify the docker container
        job = factory.create_cc3230SF_docker_job("sample_jobs/cc3220SF-docker.yaml")
        job.validate()
        description_ref = self.pipeline_reference("cc3220SF.yaml", job=job)
        self.assertEqual(description_ref, job.pipeline.describe(False))

        # Check BootGDBRetry action
        action = job.pipeline.actions[1].internal_pipeline.actions[0]
        self.assertEqual(action.name, "boot-gdb-retry")
        self.assertEqual(action.gdb, "gdb-multiarch")
        self.assertEqual(action.arguments, ["{ZEPHYR}"])
        self.assertEqual(len(action.commands), 4)
        self.assertEqual(
            action.commands[0],
            'target remote | openocd -c "gdb_port pipe" -f {OPENOCD_SCRIPT}',
        )
        self.assertEqual(action.commands[1], "monitor reset halt")
        self.assertEqual(action.commands[2], "load")
        self.assertEqual(action.commands[3], "set remotetimeout 10000")
        self.assertEqual(action.container, "ti-openocd-small")
        self.assertEqual(action.devices, ["/dev/hidraw3"])

        # use the default docker container
        job = factory.create_cc3230SF_docker_job("sample_jobs/cc3220SF.yaml")
        job.validate()
        description_ref = self.pipeline_reference("cc3220SF.yaml", job=job)
        self.assertEqual(description_ref, job.pipeline.describe(False))

        # Check BootGDBRetry action
        action = job.pipeline.actions[1].internal_pipeline.actions[0]
        self.assertEqual(action.name, "boot-gdb-retry")
        self.assertEqual(action.gdb, "gdb-multiarch")
        self.assertEqual(action.arguments, ["{ZEPHYR}"])
        self.assertEqual(len(action.commands), 4)
        self.assertEqual(
            action.commands[0],
            'target remote | openocd -c "gdb_port pipe" -f {OPENOCD_SCRIPT}',
        )
        self.assertEqual(action.commands[1], "monitor reset halt")
        self.assertEqual(action.commands[2], "load")
        self.assertEqual(action.commands[3], "set remotetimeout 10000")
        self.assertEqual(action.container, "ti-openocd")
        self.assertEqual(action.devices, ["/dev/hidraw3"])



# -*- coding: utf-8 -*-
# Copyright (C) 2015-2018 Linaro Limited
#
# Author: Neil Williams <neil.williams@linaro.org>
#         Stevan Radakovic <stevan.radakovic@linaro.org>
#
# This file is part of LAVA.
#
# LAVA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation
#
# LAVA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with LAVA.  If not, see <http://www.gnu.org/licenses/>.

from django import template


register = template.Library()


@register.simple_tag
def is_accessible_by(record, user):
    try:
        return record.is_accessible_by(user)
    except (TypeError, AttributeError, IndexError):
        return False


@register.filter
def check_chart_access(record, user):
    access = True
    for query in record.queries.all():
        if not query.is_accessible_by(user):
            access = False
    return access


@register.simple_tag
def get_extra_source(record, data):
    if not data:
        return ""
    if record.id in data:
        return data[record.id]
    return ""



# Copyright (C) 2014 Linaro Limited
#
# Author: Neil Williams <neil.williams@linaro.org>
#
# This file is part of LAVA Dispatcher.
#
# LAVA Dispatcher is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# LAVA Dispatcher is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along
# with this program; if not, see <http://www.gnu.org/licenses>.

from lava_dispatcher.action import Action
from lava_common.exceptions import LAVATimeoutError
from lava_dispatcher.logical import LavaTest, RetryAction


class TestAction(Action):
    """
    Base class for all actions which run lava test
    cases on a device under test.
    The subclass selected to do the work will be the
    subclass returning True in the accepts(device, image)
    function.
    Each new subclass needs a unit test to ensure it is
    reliably selected for the correct deployment and not
    selected for an invalid deployment or a deployment
    accepted by a different subclass.
    """

    name = "test"
    timeout_exception = LAVATimeoutError



# -*- coding: utf-8 -*-
# Generated by Django 1.10 on 2017-05-19 14:45
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("lava_scheduler_app", "0027_device_dict_onto_filesystem")]

    operations = [
        migrations.AddField(
            model_name="notification",
            name="callback_content_type",
            field=models.IntegerField(
                blank=True,
                choices=[(0, "urlencoded"), (1, "json")],
                default=None,
                null=True,
                verbose_name="Callback content-type",
            ),
        ),
        migrations.AddField(
            model_name="notification",
            name="callback_dataset",
            field=models.IntegerField(
                blank=True,
                choices=[(0, "minimal"), (1, "logs"), (2, "results"), (3, "all")],
                default=None,
                null=True,
                verbose_name="Callback dataset",
            ),
        ),
        migrations.AddField(
            model_name="notification",
            name="callback_method",
            field=models.IntegerField(
                blank=True,
                choices=[(0, "GET"), (1, "POST")],
                default=None,
                null=True,
                verbose_name="Callback method",
            ),
        ),
        migrations.AddField(
            model_name="notification",
            name="callback_token",
            field=models.CharField(
                blank=True,
                default=None,
                max_length=200,
                null=True,
                verbose_name="Callback token",
            ),
        ),
        migrations.AddField(
            model_name="notification",
            name="callback_url",
            field=models.CharField(
                blank=True,
                default=None,
                max_length=200,
                null=True,
                verbose_name="Callback URL",
            ),
        ),
    ]



# Copyright (C) 2017 Linaro Limited
#
# Author: Senthil Kumaran S <senthil.kumaran@linaro.org>
#
# This file is part of LAVA Dispatcher.
#
# LAVA Dispatcher is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# LAVA Dispatcher is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along
# with this program; if not, see <http://www.gnu.org/licenses>.

import pyudev
import time
from lava_dispatcher.action import Action
from lava_common.exceptions import LAVABug, InfrastructureError


class WaitUSBSerialDeviceAction(Action):

    name = "wait-usb-serial"
    description = "wait for USB serial device"
    summary = "wait for USB serial device"

    def __init__(self):
        super().__init__()
        self.serial_device = {}
        self.usb_sleep = 0

    def validate(self):
        super().validate()
        board_id = self.job.device.get("board_id", "")
        usb_vendor_id = self.job.device.get("usb_vendor_id", "")
        usb_product_id = self.job.device.get("usb_product_id", "")
        usb_serial_driver = self.job.device.get("usb_serial_driver", "cdc_acm")
        if board_id == "0000000000":
            self.errors = "[USBSERIAL] board_id unset"
        if usb_vendor_id == "0000":
            self.errors = "usb_vendor_id unset"
        if usb_product_id == "0000":
            self.errors = "usb_product_id unset"
        self.serial_device = {
            "ID_SERIAL_SHORT": str(board_id),
            "ID_VENDOR_ID": str(usb_vendor_id),
            "ID_MODEL_ID": str(usb_product_id),
            "ID_USB_DRIVER": str(usb_serial_driver),
        }
        self.usb_sleep = self.job.device.get("usb_sleep", 0)
        if not isinstance(self.usb_sleep, int):
            self.errors = "usb_sleep should be an integer"

    def run(self, connection, max_end_time):
        connection = super().run(connection, max_end_time)
        self.logger.debug("Waiting for usb serial device: %s", self.serial_device)
        wait_udev_event(action="add", match_dict=self.serial_device, subsystem="tty")
        if self.usb_sleep:
            self.logger.debug(
                "Waiting for the board to setup, sleeping %ds", self.usb_sleep
            )
            time.sleep(self.usb_sleep)
        return connection


class WaitDFUDeviceAction(Action):

    name = "wait-dfu-device"
    description = "wait for DFU device"
    summary = "wait for DFU device"

    def __init__(self):
        super().__init__()
        self.dfu_device = {}

    def validate(self):
        super().validate()
        board_id = self.job.device.get("board_id", "")
        usb_vendor_id = self.job.device.get("usb_vendor_id", "")
        usb_product_id = self.job.device.get("usb_product_id", "")
        if board_id == "0000000000":
            self.errors = "[DFU] board_id unset"
        if usb_vendor_id == "0000":
            self.errors = "usb_vendor_id unset"
        if usb_product_id == "0000":
            self.errors = "usb_product_id unset"
        self.dfu_device = {
            "ID_SERIAL_SHORT": str(board_id),
            "ID_VENDOR_ID": str(usb_vendor_id),
            "ID_MODEL_ID": str(usb_product_id),
        }

    def run(self, connection, max_end_time):
        connection = super().run(connection, max_end_time)
        self.logger.debug("Waiting for DFU device: %s", self.dfu_device)
        wait_udev_event(
            action="add",
            match_dict=self.dfu_device,
            subsystem="usb",
            devtype="usb_device",
        )
        return connection


class WaitUSBMassStorageDeviceAction(Action):

    name = "wait-usb-mass-storage-device"
    description = "wait for USB mass storage device"
    summary = "wait for USB mass storage device"

    def __init__(self):
        super().__init__()
        self.ms_device = {}

    def validate(self):
        super().validate()
        usb_fs_label = self.job.device.get("usb_filesystem_label")
        if not isinstance(usb_fs_label, str):
            self.errors = "usb_fs_label unset"
        self.ms_device = {"ID_FS_LABEL": str(usb_fs_label)}

    def run(self, connection, max_end_time):
        connection = super().run(connection, max_end_time)
        self.logger.debug("Waiting for USB mass storage device: %s", self.ms_device)
        wait_udev_event(
            action="add",
            match_dict=self.ms_device,
            subsystem="block",
            devtype="partition",
        )
        return connection


class WaitDevicePathAction(Action):

    name = "wait-device-path"
    description = "wait for udev device path"
    summary = "wait for udev device path"

    def __init__(self, path=None):
        super().__init__()
        self.devicepath = path

    def validate(self):
        super().validate()
        if not isinstance(self.devicepath, str):
            self.errors = "invalid device path"

    def run(self, connection, max_end_time):
        connection = super().run(connection, max_end_time)
        self.logger.debug("Waiting for udev device path: %s", self.devicepath)
        wait_udev_event(action="add", devicepath=self.devicepath)
        return connection


class WaitDeviceBoardID(Action):

    name = "wait-device-boardid"
    description = "wait for udev device with board ID"
    summary = "wait for udev device with board ID"

    def __init__(self, board_id=None):
        super().__init__()
        self.udev_device = None
        if not board_id:
            self.board_id = self.job.device.get("board_id")
        else:
            self.board_id = board_id

    def validate(self):
        super().validate()
        if not isinstance(self.board_id, str):
            self.errors = "invalid board_id"
        self.udev_device = {"ID_SERIAL_SHORT": str(self.board_id)}

    def run(self, connection, max_end_time):
        connection = super().run(connection, max_end_time)
        self.logger.debug("Waiting for udev device with ID: %s", self.board_id)
        wait_udev_event(action="add", match_dict=self.udev_device)
        return connection


def _dict_compare(d1, d2):  # pylint: disable=invalid-name
    d1_keys = set(d1.keys())
    d2_keys = set(d2.keys())
    intersect_keys = d1_keys.intersection(d2_keys)
    return set(o for o in intersect_keys if d1[o] == d2[o])


def wait_udev_event(
    action="add", match_dict=None, subsystem=None, devtype=None, devicepath=None
):
    if action not in ["add", "remove", "change"]:
        raise LAVABug(
            "Invalid action for udev to wait for: %s, expected 'add' or 'remove'"
            % action
        )
    if match_dict:
        if not isinstance(match_dict, dict):
            raise LAVABug("match_dict was not a dict")
    else:
        if devicepath:
            if not isinstance(devicepath, str):
                raise LAVABug("devicepath was not a string")
            match_dict = {}
        else:
            raise LAVABug("Neither match_dict nor devicepath were set")
    if devtype and not subsystem:
        raise LAVABug("Cannot filter udev by devtype without a subsystem")
    match_dict["ACTION"] = action
    context = pyudev.Context()
    monitor = pyudev.Monitor.from_netlink(context)
    if devtype and subsystem:
        monitor.filter_by(subsystem, devtype)
    else:
        if subsystem:
            monitor.filter_by(subsystem)
    for device in iter(monitor.poll, None):
        same = _dict_compare(dict(device), match_dict)
        if same == set(match_dict.keys()):
            if devicepath:
                if devicepath in dict(device).get("DEVLINKS", "") or devicepath in dict(
                    device
                ).get("DEVNAME", ""):
                    break
            else:
                break


def get_udev_devices(job=None, logger=None, device_info=None):
    """
    Get udev device nodes based on serial, vendor and product ID
    All subsystems are allowed so that additional hardware like
    tty devices can be added to the LXC. The ID to match is controlled
    by the lab admin.
    """
    context = pyudev.Context()
    device_paths = set()
    devices = []
    if job:
        devices = job.device.get("device_info", [])
    # device_info argument overrides job device_info
    if device_info:
        devices = device_info
    if not devices:
        return []
    added = set()
    for usb_device in devices:
        board_id = str(usb_device.get("board_id", ""))
        usb_vendor_id = str(usb_device.get("usb_vendor_id", ""))
        usb_product_id = str(usb_device.get("usb_product_id", ""))
        usb_fs_label = str(usb_device.get("fs_label", ""))
        # check if device is already connected
        # try with all parameters such as board id, usb_vendor_id and
        # usb_product_id
        for device in context.list_devices():
            if board_id and usb_vendor_id and usb_product_id:
                if (
                    (device.get("ID_SERIAL_SHORT") == board_id)
                    and (device.get("ID_VENDOR_ID") == usb_vendor_id)
                    and (device.get("ID_MODEL_ID") == usb_product_id)
                ):
                    device_paths.add(device.device_node)
                    added.add(board_id)
                    for child in device.children:
                        if child.device_node:
                            device_paths.add(child.device_node)
                    for link in device.device_links:
                        device_paths.add(link)
            elif board_id and usb_vendor_id and not usb_product_id:
                # try with parameters such as board id, usb_vendor_id
                if (device.get("ID_SERIAL_SHORT") == board_id) and (
                    device.get("ID_VENDOR_ID") == usb_vendor_id
                ):
                    device_paths.add(device.device_node)
                    added.add(board_id)
                    for child in device.children:
                        if child.device_node:
                            device_paths.add(child.device_node)
                    for link in device.device_links:
                        device_paths.add(link)
            elif board_id and not usb_vendor_id and not usb_product_id:
                # try with board id alone
                if device.get("ID_SERIAL_SHORT") == board_id:
                    device_paths.add(device.device_node)
                    added.add(board_id)
                    for child in device.children:
                        if child.device_node:
                            device_paths.add(child.device_node)
                    for link in device.device_links:
                        device_paths.add(link)
            elif usb_fs_label:
                # Just restrict by filesystem label.
                if device.get("ID_FS_LABEL") == usb_fs_label:
                    device_paths.add(device.device_node)
                    added.add(usb_fs_label)
                    for child in device.children:
                        if child.device_node:
                            device_paths.add(child.device_node)
                    for link in device.device_links:
                        device_paths.add(link)
    if device_info:
        for static_device in device_info:
            for _, value in static_device.items():
                if value not in added:
                    raise InfrastructureError(
                        "Unable to add all static devices: board_id '%s' was not found"
                        % value
                    )
    if logger and device_paths:
        logger.debug("Adding %s", ", ".join(device_paths))
    return list(device_paths)


def usb_device_wait(job, device_actions=None):
    context = pyudev.Context()
    if device_actions is None:
        device_actions = []
    for usb_device in job.device.get("device_info", []):
        board_id = str(usb_device.get("board_id", ""))
        usb_vendor_id = str(usb_device.get("usb_vendor_id", ""))
        usb_product_id = str(usb_device.get("usb_product_id", ""))
        usb_fs_label = str(usb_device.get("fs_label", ""))
        # monitor for device
        monitor = pyudev.Monitor.from_netlink(context)
        if board_id and usb_vendor_id and usb_product_id:
            for device in iter(monitor.poll, None):
                if (
                    (device.get("ID_SERIAL_SHORT") == board_id)
                    and (device.get("ID_VENDOR_ID") == usb_vendor_id)
                    and (device.get("ID_MODEL_ID") == usb_product_id)
                    and device.action in device_actions
                ):
                    break
            return
        elif board_id and usb_vendor_id and not usb_product_id:
            for device in iter(monitor.poll, None):
                if (
                    (device.get("ID_SERIAL_SHORT") == board_id)
                    and (device.get("ID_VENDOR_ID") == usb_vendor_id)
                    and device.action in device_actions
                ):
                    break
            return
        elif board_id and not usb_vendor_id and not usb_product_id:
            for device in iter(monitor.poll, None):
                if (
                    device.get("ID_SERIAL_SHORT") == board_id
                ) and device.action in device_actions:
                    break
            return
        elif usb_fs_label:
            for device in iter(monitor.poll, None):
                if (
                    device.get("ID_FS_LABEL") == usb_fs_label
                ) and device.action in device_actions:
                    break
            return


def allow_fs_label(device):
    # boot/deploy methods that indicate that the device in question
    # will require a filesystem label to identify a device.
    # So far, mps devices are supported, but these don't provide a
    # unique serial, so fs label must be used.
    fs_label_methods = ["mps", "recovery"]

    # Don't allow using filesystem labels by default as they are
    # unreliable, and can be changed via a malicious job.
    _allow_fs_label = False
    for method in fs_label_methods:
        if (
            method in device["actions"]["boot"]["methods"]
            or method in device["actions"]["deploy"]["methods"]
        ):
            _allow_fs_label = True
    return _allow_fs_label


def lxc_udev_rule(data):
    """Construct the udev rule string."""
    rule = 'ACTION=="add", '
    if data["serial_number"]:
        rule += 'ATTR{{serial}}=="{serial_number}", '
    if data["vendor_id"] is not None:
        rule += 'ATTR{{idVendor}}=="{vendor_id}", '
    if data["product_id"] is not None:
        rule += 'ATTR{{idProduct}}=="{product_id}", '
    if data["fs_label"] is not None:
        rule += 'ENV{{ID_FS_LABEL}}=="{fs_label}", '
    rule += (
        'RUN+="/usr/share/lava-dispatcher/lava_lxc_device_add.py '
        "--lxc-name {lxc_name} --device-node $name "
        "--job-id {job_id}"
    )
    if data["logging_url"]:
        rule += " --logging-url {logging_url}"
    rule = rule.format(**data)

    if data["master_cert"] is not None:
        rule += " --master-cert %s" % data["master_cert"]
    if data["slave_cert"] is not None:
        rule += " --slave-cert %s" % data["slave_cert"]
    if data["ipv6"]:
        rule += " --ipv6"
    if data["fs_label"] is not None:
        rule = 'IMPORT{builtin}="blkid"\n' + rule
    rule += '"\n'
    return rule


def lxc_udev_rule_parent(data):
    """Construct the udev rule string."""
    rule = 'ACTION=="add", '
    if data["vendor_id"] is not None:
        rule += 'ATTRS{{idVendor}}=="{vendor_id}", '
    if data["product_id"] is not None:
        rule += 'ATTRS{{idProduct}}=="{product_id}", '
    if data["fs_label"] is not None:
        rule += 'ENV{{ID_FS_LABEL}}=="{fs_label}", '
    rule += (
        'RUN+="/usr/share/lava-dispatcher/lava_lxc_device_add.py '
        "--lxc-name {lxc_name} --device-node $name "
        "--job-id {job_id}"
    )
    if data["logging_url"]:
        rule += " --logging-url {logging_url}"
    rule = rule.format(**data)

    if data["master_cert"] is not None:
        rule += " --master-cert %s" % data["master_cert"]
    if data["slave_cert"] is not None:
        rule += " --slave-cert %s" % data["slave_cert"]
    if data["ipv6"]:
        rule += " --ipv6"
    if data["fs_label"] is not None:
        rule = 'IMPORT{builtin}="blkid"\n' + rule
    rule += '"\n'
    return rule



# coding: utf-8
#
# Copyright (C) 2018 Linaro Limited
#
# Author: Rémi Duraffort <remi.duraffort@linaro.org>
#
# This file is part of LAVA.
#
# LAVA is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# LAVA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along
# with this program; if not, see <http://www.gnu.org/licenses>.

from voluptuous import Any, Match, Optional, Required

from lava_common.schemas import test


def schema(live=False):
    common = {
        Required("path"): str,
        Required("name"): str,
        Optional("skip_install"): [
            Any("keys", "sources", "deps", "steps", "git-repos", "all")
        ],
        # Optional("parameters"):
    }

    base = {
        Required("definitions"): [
            Any(
                {
                    Required("repository"): str,
                    Required("from"): Any("git", "bzr"),
                    Optional("branch"): str,
                    Optional("revision"): str,
                    Optional("parameters"): {str: str},
                    Optional("params"): {str: str},
                    **common,
                },
                {
                    Required("repository"): {
                        Required("metadata"): dict,  # TODO: what's required?
                        Optional("install"): dict,  # TODO: check
                        Required("run"): {Required("steps"): [str]},
                    },
                    Required("from"): "inline",
                    **common,
                },
            )
        ]
    }
    return {**test.schema(live), **base}



# Copyright (C) 2014 Linaro Limited
#
# Author: Neil Williams <neil.williams@linaro.org>
#
# This file is part of LAVA Dispatcher.
#
# LAVA Dispatcher is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# LAVA Dispatcher is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along
# with this program; if not, see <http://www.gnu.org/licenses>.

import yaml
from yaml.composer import Composer
from yaml.constructor import Constructor
from lava_dispatcher.job import Job
from lava_dispatcher.action import Pipeline, Timeout, JobError
from lava_dispatcher.logical import Deployment, Boot, LavaTest
from lava_dispatcher.deployment_data import get_deployment_data
from lava_dispatcher.power import FinalizeAction
from lava_dispatcher.connection import Protocol

# Bring in the strategy subclass lists, ignore pylint warnings.
# pylint: disable=unused-import,too-many-arguments,too-many-nested-blocks,too-many-branches
from lava_dispatcher.actions.commands import CommandAction
import lava_dispatcher.actions.deploy.strategies
import lava_dispatcher.actions.boot.strategies
import lava_dispatcher.actions.test.strategies
import lava_dispatcher.protocols.strategies


def parse_action(job_data, name, device, pipeline, test_info, test_count):
    """
    If protocols are defined, each Action may need to be aware of the protocol parameters.
    """
    parameters = job_data[name]
    if "protocols" in pipeline.job.parameters:
        parameters.update(pipeline.job.parameters["protocols"])

    if name == "boot":
        Boot.select(device, parameters)(pipeline, parameters)
    elif name == "test":
        # stage starts at 0
        parameters["stage"] = test_count - 1
        action = LavaTest.select(device, parameters)
        action(pipeline, parameters)
        return action
    elif name == "deploy":
        candidate = Deployment.select(device, parameters)
        if parameters["namespace"] in test_info and candidate.uses_deployment_data():
            if any(
                [
                    testclass
                    for testclass in test_info[parameters["namespace"]]
                    if testclass["class"].needs_deployment_data()
                ]
            ):
                parameters.update(
                    {"deployment_data": get_deployment_data(parameters.get("os", ""))}
                )
        if "preseed" in parameters:
            parameters.update(
                {"deployment_data": get_deployment_data(parameters.get("os", ""))}
            )
        Deployment.select(device, parameters)(pipeline, parameters)


class JobParser:
    """
    Creates a Job object from the Device and the job YAML by selecting the
    Strategy class with the highest priority for the parameters of the job.

    Adding new behaviour is a two step process:
     - always add a new Action, usually with an internal pipeline, to implement the new behaviour
     - add a new Strategy class which creates a suitable pipeline to use that Action.

    Re-use existing Action classes wherever these can be used without changes.

    If two or more Action classes have very similar behaviour, re-factor to make a
    new base class for the common behaviour and retain the specialised classes.

    Strategy selection via select() must only ever rely on the device and the
    job parameters. Add new parameters to the job to distinguish strategies, e.g.
    the boot method or deployment method.
    """

    # FIXME: needs a Schema and a check routine

    def _timeouts(self, data, job):  # pylint: disable=no-self-use
        if "job" in data.get("timeouts", {}):
            duration = Timeout.parse(data["timeouts"]["job"])
            job.timeout = Timeout("job", duration)

    # pylint: disable=too-many-locals,too-many-statements
    def parse(self, content, device, job_id, logger, dispatcher_config, env_dut=None):
        data = yaml.safe_load(content)
        job = Job(job_id, data, logger)
        test_counts = {}
        job.device = device
        job.parameters["env_dut"] = env_dut
        # Load the dispatcher config
        job.parameters["dispatcher"] = {}
        if dispatcher_config is not None:
            config = yaml.safe_load(dispatcher_config)
            if isinstance(config, dict):
                job.parameters["dispatcher"] = config

        level_tuple = Protocol.select_all(job.parameters)
        # sort the list of protocol objects by the protocol class level.
        job.protocols = [
            item[0](job.parameters, job_id)
            for item in sorted(level_tuple, key=lambda level_tuple: level_tuple[1])
        ]
        pipeline = Pipeline(job=job)
        self._timeouts(data, job)

        # deploy and boot classes can populate the pipeline differently depending
        # on the test action type they are linked with (via namespacing).
        # This code builds an information dict for each namespace which is then
        # passed as a parameter to each Action class to use.
        test_actions = [action for action in data["actions"] if "test" in action]
        for test_action in test_actions:
            test_parameters = test_action["test"]
            test_type = LavaTest.select(device, test_parameters)
            namespace = test_parameters.get("namespace", "common")
            connection_namespace = test_parameters.get(
                "connection-namespace", namespace
            )
            if namespace in job.test_info:
                job.test_info[namespace].append(
                    {"class": test_type, "parameters": test_parameters}
                )
            else:
                job.test_info.update(
                    {namespace: [{"class": test_type, "parameters": test_parameters}]}
                )
            if namespace != connection_namespace:
                job.test_info.update(
                    {
                        connection_namespace: [
                            {"class": test_type, "parameters": test_parameters}
                        ]
                    }
                )

        # FIXME: also read permissable overrides from device config and set from job data
        # FIXME: ensure that a timeout for deployment 0 does not get set as the timeout for deployment 1 if 1 is default
        for action_data in data["actions"]:
            for name in action_data:
                # Set a default namespace if needed
                namespace = action_data[name].setdefault("namespace", "common")
                test_counts.setdefault(namespace, 1)

                if name == "deploy" or name == "boot" or name == "test":
                    action = parse_action(
                        action_data,
                        name,
                        device,
                        pipeline,
                        job.test_info,
                        test_counts[namespace],
                    )
                    if name == "test" and action.needs_overlay():
                        test_counts[namespace] += 1
                elif name == "repeat":
                    count = action_data[name][
                        "count"
                    ]  # first list entry must be the count dict
                    repeats = action_data[name]["actions"]
                    for c_iter in range(count):
                        for repeating in repeats:  # block of YAML to repeat
                            for (
                                repeat_action
                            ) in repeating:  # name of the action for this block
                                repeating[repeat_action]["repeat-count"] = c_iter
                                namespace = repeating[repeat_action].setdefault(
                                    "namespace", "common"
                                )
                                test_counts.setdefault(namespace, 1)
                                action = parse_action(
                                    repeating,
                                    repeat_action,
                                    device,
                                    pipeline,
                                    job.test_info,
                                    test_counts[namespace],
                                )
                                if repeat_action == "test" and action.needs_overlay():
                                    test_counts[namespace] += 1

                elif name == "command":
                    action = CommandAction()
                    action.parameters = action_data[name]
                    pipeline.add_action(action)

                else:
                    raise JobError("Unknown action name '%s'" % name)

        # there's always going to need to be a finalize_process action
        finalize = FinalizeAction()
        pipeline.add_action(finalize)
        finalize.populate(None)
        job.pipeline = pipeline
        if "compatibility" in data:
            try:
                job_c = int(job.compatibility)
                data_c = int(data["compatibility"])
            except ValueError as exc:
                raise JobError("invalid compatibility value: %s" % exc)
            if job_c < data_c:
                raise JobError(
                    "Dispatcher unable to meet job compatibility requirement. %d > %d"
                    % (job_c, data_c)
                )
        return job



# Copyright (C) 2015 Linaro Limited
#
# Author: Neil Williams <neil.williams@linaro.org>
#
# This file is part of LAVA Dispatcher.
#
# LAVA Dispatcher is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# LAVA Dispatcher is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along
# with this program; if not, see <http://www.gnu.org/licenses>.

# pylint: disable=too-many-return-statements,too-many-instance-attributes

import os
import yaml
from lava_common.exceptions import LAVABug, JobError
from lava_dispatcher.action import Action, Pipeline
from lava_dispatcher.logical import Boot, RetryAction
from lava_dispatcher.actions.boot import AutoLoginAction
from lava_dispatcher.actions.boot.environment import ExportDeviceEnvironment
from lava_dispatcher.utils.shell import which
from lava_dispatcher.shell import ExpectShellSession
from lava_dispatcher.connections.ssh import ConnectSsh
from lava_dispatcher.protocols.multinode import MultinodeProtocol


class SshLogin(Boot):
    """
    Ssh boot strategy is a login process, without actually booting a kernel
    but still needs AutoLoginAction.
    """

    compatibility = 1

    def __init__(self, parent, parameters):
        super().__init__(parent)
        self.action = SshAction()
        self.action.job = self.job
        parent.add_action(self.action, parameters)

    @classmethod
    def accepts(cls, device, parameters):
        if "ssh" not in device["actions"]["boot"]["methods"]:
            return False, '"ssh" not in device configuration boot methods'
        if "ssh" not in parameters["method"]:
            return False, '"ssh" not in "method"'
        return True, "accepted"


class SshAction(RetryAction):
    """
    Simple action to wrap AutoLoginAction and ExpectShellSession
    """

    section = "boot"
    name = "login-ssh"
    description = "connect over ssh and ensure a shell is found"
    summary = "login over ssh"

    def populate(self, parameters):
        self.internal_pipeline = Pipeline(
            parent=self, job=self.job, parameters=parameters
        )
        scp = Scp("overlay")
        self.internal_pipeline.add_action(scp)
        self.internal_pipeline.add_action(PrepareSsh())
        self.internal_pipeline.add_action(ConnectSsh())
        self.internal_pipeline.add_action(AutoLoginAction())
        self.internal_pipeline.add_action(ExpectShellSession())
        self.internal_pipeline.add_action(ExportDeviceEnvironment())
        self.internal_pipeline.add_action(ScpOverlayUnpack())


class Scp(ConnectSsh):
    """
    Use the SSH connection options to copy files over SSH
    One action per scp operation, just as with download action
    Needs the reference into the common data for each file to copy
    This is a Deploy action. lava-start is managed by the protocol,
    when this action starts, the device is in the "receiving" state.
    """

    name = "scp-deploy"
    description = "copy a file to a known device using scp"
    summary = "scp over the ssh connection"

    def __init__(self, key):
        super().__init__()
        self.key = key
        self.scp = []

    def validate(self):
        super().validate()
        params = self._check_params()
        which("scp")
        if "ssh" not in self.job.device["actions"]["deploy"]["methods"]:
            self.errors = "Unable to use %s without ssh deployment" % self.name
        if "ssh" not in self.job.device["actions"]["boot"]["methods"]:
            self.errors = "Unable to use %s without ssh boot" % self.name
        if self.get_namespace_data(
            action="prepare-scp-overlay", label="prepare-scp-overlay", key=self.key
        ):
            self.primary = False
        elif "host" not in self.job.device["actions"]["deploy"]["methods"]["ssh"]:
            self.errors = "Invalid device or job configuration, missing host."
        if (
            not self.primary
            and len(
                self.get_namespace_data(
                    action="prepare-scp-overlay",
                    label="prepare-scp-overlay",
                    key=self.key,
                )
            )
            != 1
        ):
            self.errors = "Invalid number of host_keys"
        if self.primary:
            host_address = self.job.device["actions"]["deploy"]["methods"]["ssh"][
                "host"
            ]
            if not host_address:
                self.errors = (
                    "Unable to retrieve ssh_host address for primary connection."
                )
        if "port" in self.job.device["actions"]["deploy"]["methods"]["ssh"]:
            port = str(self.job.device["actions"]["deploy"]["methods"]["ssh"]["port"])
            if not port.isdigit():
                self.errors = "Port was set but was not a digit"
        if self.valid:
            self.scp.append("scp")
            if "options" in params:
                self.scp.extend(params["options"])

    def run(self, connection, max_end_time):
        path = self.get_namespace_data(
            action="prepare-scp-overlay", label="scp-deploy", key=self.key
        )
        if not path:
            error_msg = "%s: could not find details of '%s'" % (self.name, self.key)
            self.logger.error(error_msg)
            raise JobError(error_msg)

        overrides = self.get_namespace_data(
            action="prepare-scp-overlay", label="prepare-scp-overlay", key=self.key
        )
        if self.primary:
            host_address = self.job.device["actions"]["deploy"]["methods"]["ssh"][
                "host"
            ]
        else:
            self.logger.info(
                "Retrieving common data for prepare-scp-overlay using %s",
                ",".join(overrides),
            )
            host_address = str(
                self.get_namespace_data(
                    action="prepare-scp-overlay",
                    label="prepare-scp-overlay",
                    key=overrides[0],
                )
            )
            self.logger.debug("Using common data for host: %s", host_address)
        if not host_address:
            error_msg = "%s: could not find host for deployment using %s" % (
                self.name,
                self.key,
            )
            self.logger.error(error_msg)
            raise JobError(error_msg)

        destination = "%s-%s" % (self.job.job_id, os.path.basename(path))
        command = self.scp[:]  # local copy
        # add the argument for setting the port (-P port)
        command.extend(self.scp_port)
        connection = super().run(connection, max_end_time)
        if self.identity_file:
            command.extend(["-i", self.identity_file])
        # add arguments to ignore host key checking of the host device
        command.extend(
            ["-o", "UserKnownHostsFile=/dev/null", "-o", "StrictHostKeyChecking=no"]
        )
        # add the local file as source
        command.append(path)
        command_str = " ".join(str(item) for item in command)
        self.logger.info(
            "Copying %s using %s to %s", self.key, command_str, host_address
        )
        # add the remote as destination, with :/ top level directory
        command.extend(["%s@%s:/%s" % (self.ssh_user, host_address, destination)])
        self.logger.info(yaml.dump(command))
        self.run_command(command)
        connection = super().run(connection, max_end_time)
        self.results = {"success": "ssh deployment"}
        self.set_namespace_data(
            action=self.name,
            label="scp-overlay-unpack",
            key="overlay",
            value=destination,
        )
        self.set_namespace_data(
            action="shared", label="shared", key="connection", value=connection
        )
        return connection


class PrepareSsh(Action):
    """
    Sets the host for the ConnectSsh
    """

    name = "prepare-ssh"
    description = "determine which address to use for primary or secondary connections"
    summary = "set the host address of the ssh connection"

    def __init__(self):
        super().__init__()
        self.primary = False

    def validate(self):
        if (
            "parameters" in self.parameters
            and "hostID" in self.parameters["parameters"]
        ):
            self.set_namespace_data(
                action=self.name, label="ssh-connection", key="host", value=True
            )
        else:
            self.set_namespace_data(
                action=self.name, label="ssh-connection", key="host", value=False
            )
            self.primary = True

    def run(self, connection, max_end_time):
        connection = super().run(connection, max_end_time)
        if not self.primary:
            host_data = self.get_namespace_data(
                action=MultinodeProtocol.name,
                label=MultinodeProtocol.name,
                key=self.parameters["parameters"]["hostID"],
            )
            if not host_data:
                raise JobError(
                    "Unable to retrieve %s - missing ssh deploy?"
                    % self.parameters["parameters"]["hostID"]
                )
            self.set_namespace_data(
                action=self.name,
                label="ssh-connection",
                key="host_address",
                value=host_data[self.parameters["parameters"]["host_key"]],
            )
        return connection


class ScpOverlayUnpack(Action):

    name = "scp-overlay-unpack"
    description = "unpack the overlay over an existing ssh connection"
    summary = "unpack the overlay on the remote device"

    def run(self, connection, max_end_time):
        connection = super().run(connection, max_end_time)
        if not connection:
            raise LAVABug("Cannot unpack, no connection available.")
        filename = self.get_namespace_data(
            action="scp-deploy", label="scp-overlay-unpack", key="overlay"
        )
        tar_flags = self.get_namespace_data(
            action="scp-overlay", label="scp-overlay", key="tar_flags"
        )
        cmd = "tar %s -C / -xzf /%s" % (tar_flags, filename)
        connection.sendline(cmd)
        self.wait(connection)
        self.set_namespace_data(
            action="shared", label="shared", key="connection", value=connection
        )
        return connection


class Schroot(Boot):
    def __init__(self, parent, parameters):
        super().__init__(parent)
        self.action = SchrootAction()
        self.action.job = self.job
        parent.add_action(self.action, parameters)

    @classmethod
    def accepts(cls, device, parameters):
        if "actions" not in device or "boot" not in device["actions"]:
            return False, '"boot" was not in the device configuration actions'
        if "methods" not in device["actions"]["boot"]:
            return False, '"methods" was not in the device config'
        if "schroot" not in device["actions"]["boot"]["methods"]:
            return False, '"schroot" was not in the device configuration boot methods'
        if "method" not in parameters:
            return False, '"method" was not in parameters'
        if "schroot" not in parameters["method"]:
            return False, '"method" was not "schroot"'
        return True, "accepted"


class SchrootAction(Action):
    """
    Extends the login to enter an existing schroot as a new schroot session
    using the current connection.
    Does not rely on ssh
    """

    name = "schroot-login"
    description = "enter schroot using existing connection"
    summary = "enter specified schroot"

    def __init__(self):
        super().__init__()
        self.section = "boot"
        self.schroot = None
        self.command = None

    def validate(self):
        """
        The unit test skips if schroot is not installed, the action marks the
        pipeline as invalid if schroot is not installed.
        """
        if "schroot" not in self.parameters:
            return
        if "schroot" not in self.job.device["actions"]["boot"]["methods"]:
            self.errors = "No schroot support in device boot methods"
            return
        which("schroot")
        # device parameters are for ssh
        params = self.job.device["actions"]["boot"]["methods"]
        if "command" not in params["schroot"]:
            self.errors = "Missing schroot command in device configuration"
            return
        if "name" not in params["schroot"]:
            self.errors = "Missing schroot name in device configuration"
            return
        self.schroot = params["schroot"]["name"]
        self.command = params["schroot"]["command"]

    def run(self, connection, max_end_time):
        if not connection:
            return connection
        self.logger.info("Entering %s schroot", self.schroot)
        connection.prompt_str = "(%s)" % self.schroot
        connection.sendline(self.command)
        self.wait(connection)
        # TODO: not calling super?
        return connection



# -*- coding: utf-8 -*-
#
# Copyright (C) 2017 Linaro Limited
#
# Author: Remi Duraffort <remi.duraffort@linaro.org>
#
# This file is part of LAVA Dispatcher.
#
# LAVA Dispatcher is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# LAVA Dispatcher is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses>.

from functools import wraps


def replace_exception(cls_from, cls_to, limit=2048):
    def replace_exception_wrapper(func):
        @wraps(func)
        def function_wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except cls_from as exc:
                raise cls_to(str(exc)[:limit])

        return function_wrapper

    return replace_exception_wrapper



# -*- coding: utf-8 -*-
# Copyright (C) 2017-2018 Linaro Limited
#
# Author: Remi Duraffort <remi.duraffort@linaro.org>
#
# This file is part of LAVA.
#
# LAVA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation
#
# LAVA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with LAVA.  If not, see <http://www.gnu.org/licenses/>.

import yaml
import xmlrpc.client

from django.contrib.auth.models import User, Group
from django.core.exceptions import ValidationError
from django.db import IntegrityError, transaction

from linaro_django_xmlrpc.models import ExposedV2API
from lava_scheduler_app.api import check_perm
from lava_scheduler_app.models import Device, DeviceType, Tag, Worker


class SchedulerDevicesAPI(ExposedV2API):
    @check_perm("lava_scheduler_app.add_device")
    def add(
        self,
        hostname,
        type_name,
        worker_hostname,
        user_name=None,
        group_name=None,
        public=True,
        health=None,
        description=None,
    ):
        """
        Name
        ----
        `scheduler.devices.add` (`hostname`, `type_name`, `worker_hostname`,
                                 `user_name=None`, `group_name=None`,
                                 `public=True`, `health=None`,
                                 `description=None`)

        Description
        -----------
        [superuser only]
        Add a new device to the database, to support V2.

        Each device will also need a device dictionary.

        Arguments
        ---------
        `hostname`: string
          Hostname of the device
        `type_name`: string
          Type of the new device
        `worker_hostname`: string
          Worker hostname
        `user_name`: string
          Device owner, None by default
        `group_name`: string
          Device group owner, None by default
        `public`: boolean
          Is the device public?
        `health`: string
          Device health, among ["GOOD", "UNKNOWN", "LOOPING", "BAD", "MAINTENANCE", "RETIRED"]
        `description`: string
          Device description

        Return value
        ------------
        None
        """
        user = group = None
        try:
            device_type = DeviceType.objects.get(name=type_name)
            worker = Worker.objects.get(hostname=worker_hostname)
            if user_name is not None:
                user = User.objects.get(username=user_name)
            if group_name is not None:
                group = Group.objects.get(name=group_name)
        except DeviceType.DoesNotExist:
            raise xmlrpc.client.Fault(404, "DeviceType '%s' was not found." % type_name)
        except Worker.DoesNotExist:
            raise xmlrpc.client.Fault(
                404, "Worker '%s' was not found." % worker_hostname
            )
        except User.DoesNotExist:
            raise xmlrpc.client.Fault(404, "User '%s' was not found." % user_name)
        except Group.DoesNotExist:
            raise xmlrpc.client.Fault(404, "Group '%s' was not found." % group_name)

        health_val = Device.HEALTH_UNKNOWN
        try:
            if health is not None:
                health_val = Device.HEALTH_REVERSE[health]
        except KeyError:
            raise xmlrpc.client.Fault(400, "Invalid health")

        try:
            Device.objects.create(
                hostname=hostname,
                device_type=device_type,
                user=user,
                group=group,
                is_public=public,
                state=Device.STATE_IDLE,
                health=health_val,
                worker_host=worker,
                description=description,
            )

        except (IntegrityError, ValidationError) as exc:
            raise xmlrpc.client.Fault(400, "Bad request: %s" % exc.message)

    def get_dictionary(self, hostname, render=False, context=None):
        """
        Name
        ----
        `scheduler.devices.get_dictionary` (`hostname`, `render=False`, `context=None`)

        Support for the context argument is new in api_version 2
        see system.api_version().

        Description
        -----------
        Return the device configuration

        Arguments
        ---------
        `hostname`: string
          Hostname of the device
        `render`: bool
          Render the device configuration. By default, return the dictionary
        `context`: string
          Some device templates need a context specific when processing the
          device-type template. This can be specified as a YAML string.
          New in api_version 2 - see system.api_version()

        Return value
        ------------
        The device dictionary
        """
        try:
            device = Device.objects.get(hostname=hostname)
        except Device.DoesNotExist:
            raise xmlrpc.client.Fault(404, "Device '%s' was not found." % hostname)

        if not device.is_visible_to(self.user):
            raise xmlrpc.client.Fault(
                403, "Device '%s' not available to user '%s'." % (hostname, self.user)
            )

        job_ctx = None
        if context is not None:
            try:
                job_ctx = yaml.safe_load(context)
            except yaml.YAMLError as exc:
                raise xmlrpc.client.Fault(
                    400, "Job Context '%s' is not valid: %s" % (context, str(exc))
                )

        config = device.load_configuration(
            job_ctx=job_ctx, output_format="raw" if not render else "yaml"
        )
        if config is None:
            raise xmlrpc.client.Fault(
                404, "Device '%s' does not have a configuration" % hostname
            )
        return xmlrpc.client.Binary(config.encode("utf-8"))

    @check_perm("lava_scheduler_app.change_device")
    def set_dictionary(self, hostname, dictionary):
        """
        Name
        ----
        `scheduler.devices.set_dictionary` (`hostname`, `dictionary`)

        Description
        -----------
        [superuser only]
        Set the device dictionary

        Arguments
        ---------
        `hostname`: string
          Hostname of the device
        `dictionary`: string
          The device dictionary as a jinja2 template

        Return value
        ------------
        True if the dictionary was saved to file, False otherwise.
        """
        try:
            device = Device.objects.get(hostname=hostname)
        except Device.DoesNotExist:
            raise xmlrpc.client.Fault(404, "Device '%s' was not found." % hostname)

        return device.save_configuration(dictionary)

    def list(self, show_all=False, offline_info=False):
        """
        Name
        ----
        `scheduler.devices.list` (`show_all=False`, `offline_info=False`)

        Description
        -----------
        List available devices with their state and type information.

        Arguments
        ---------
        `show_all`: boolean
          Show all devices, including retired
        `offline_info`: boolean
          Add date from which each of the returned devices is offline (if the
          device is offline) and the user who put the device offline (if the
          device is offline) to the returned dictionary.

        Return value
        ------------
        This function returns an XML-RPC array in which each item is a
        dictionary with device information
        """
        devices = Device.objects.all().select_related("device_type")
        if not show_all:
            devices = devices.exclude(health=Device.HEALTH_RETIRED)
        devices = devices.order_by("hostname")

        ret = []
        for device in devices:
            if device.is_visible_to(self.user):
                current_job = device.current_job()
                device_dict = {
                    "hostname": device.hostname,
                    "type": device.device_type.name,
                    "health": device.get_health_display(),
                    "state": device.get_state_display(),
                    "current_job": current_job.pk if current_job else None,
                    "pipeline": True,
                }
                ret.append(device_dict)

        return ret

    def show(self, hostname):
        """
        Name
        ----
        `scheduler.devices.show` (`hostname`)

        Description
        -----------
        Show some details about the given device.

        Arguments
        ---------
        `hostname`: string
          Hostname of the device

        Return value
        ------------
        This function returns an XML-RPC dictionary with device details
        """
        try:
            device = Device.objects.get(hostname=hostname)
        except Device.DoesNotExist:
            raise xmlrpc.client.Fault(404, "Device '%s' was not found." % hostname)

        if not device.is_visible_to(self.user):
            raise xmlrpc.client.Fault(
                403, "Device '%s' not available to user '%s'." % (hostname, self.user)
            )

        current_job = device.current_job()
        device_dict = {
            "hostname": device.hostname,
            "device_type": device.device_type.name,
            "health": device.get_health_display(),
            "state": device.get_state_display(),
            "health_job": bool(device.get_health_check()),
            "description": device.description,
            "public": device.is_public,
            "pipeline": True,
            "has_device_dict": bool(device.load_configuration(output_format="raw")),
            "worker": None,
            "user": device.user.username if device.user else None,
            "group": device.group.name if device.group else None,
            "current_job": current_job.pk if current_job else None,
            "tags": [t.name for t in device.tags.all().order_by("name")],
        }
        if device.worker_host is not None:
            device_dict["worker"] = device.worker_host.hostname

        return device_dict

    @check_perm("lava_scheduler_app.change_device")
    def update(
        self,
        hostname,
        worker_hostname=None,
        user_name=None,
        group_name=None,
        public=True,
        health=None,
        description=None,
    ):
        """
        Name
        ----
        `scheduler.devices.update` (`hostname`, `worker_hostname=None`,
                                    `user_name=None`, `group_name=None`,
                                    `public=True`, `health=None`,
                                    `description=None`)

        Description
        -----------
        [superuser only]
        Update device parameters. Only the non-None values will be updated.
        Owner and group are always updated at the same time.

        Arguments
        ---------
        `hostname`: string
          Hostname of the device
        `worker_hostname`: string
          Worker hostname
        `user_name`: string
          Device owner
        `group_name`: string
          Device group owner
        `public`: boolean
          Is the device public?
        `health`: string
          Device health, among ["GOOD", "UNKNOWN", "LOOPING", "BAD", "MAINTENANCE", "RETIRED"]
        `description`: string
          Device description

        Return value
        ------------
        None
        """
        try:
            with transaction.atomic():
                device = Device.objects.get(hostname=hostname)

                if worker_hostname is not None:
                    try:
                        device.worker_host = Worker.objects.get(
                            hostname=worker_hostname
                        )
                    except Worker.DoesNotExist:
                        raise xmlrpc.client.Fault(
                            404, "Unable to find worker '%s'" % worker_hostname
                        )

                user = group = None
                try:
                    if user_name is not None:
                        user = User.objects.get(username=user_name)
                    if group_name is not None:
                        group = Group.objects.get(name=group_name)
                except User.DoesNotExist:
                    raise xmlrpc.client.Fault(
                        404, "User '%s' was not found." % user_name
                    )
                except Group.DoesNotExist:
                    raise xmlrpc.client.Fault(
                        404, "Group '%s' was not found." % group_name
                    )

                if user is not None or group is not None:
                    device.user = user
                    device.group = group

                if public is not None:
                    device.is_public = public

                try:
                    if health is not None:
                        prev_health = device.get_health_display()
                        device.health = Device.HEALTH_REVERSE[health]
                        device.log_admin_entry(
                            self.user,
                            "%s → %s (xmlrpc api)"
                            % (prev_health, device.get_health_display()),
                        )
                except KeyError:
                    raise xmlrpc.client.Fault(400, "Health '%s' is invalid" % health)

                if description is not None:
                    device.description = description

                device.save()
        except Device.DoesNotExist:
            raise xmlrpc.client.Fault(404, "Device '%s' was not found." % hostname)
        except (IntegrityError, ValidationError) as exc:
            raise xmlrpc.client.Fault(400, "Bad request: %s" % exc.message)


class SchedulerDevicesTagsAPI(ExposedV2API):
    @check_perm("lava_scheduler_app.add_tag")
    @check_perm("lava_scheduler_app.change_device")
    def add(self, hostname, name):
        """
        Name
        ----
        `scheduler.devices.tags.add` (`hostname`, `name`)

        Description
        -----------
        [superuser only]
        Add a device tag to the specific device

        Arguments
        ---------
        `hostname`: string
          Device hostname
        `name`: string
          Tag name to add

        Return value
        ------------
        None
        """
        try:
            device = Device.objects.get(hostname=hostname)
        except Device.DoesNotExist:
            raise xmlrpc.client.Fault(404, "Device '%s' was not found." % hostname)

        tag, _ = Tag.objects.get_or_create(name=name)
        device.tags.add(tag)
        device.save()

    def list(self, hostname):
        """
        Name
        ----
        `scheduler.devices.tags.list` (`hostname`)

        Description
        -----------
        List device tags

        Arguments
        ---------
        `hostname`: string
          Device hostname

        Return value
        ------------
        This function returns an XML-RPC array of tag names
        """
        try:
            device = Device.objects.get(hostname=hostname)
        except Device.DoesNotExist:
            raise xmlrpc.client.Fault(404, "Device '%s' was not found." % hostname)

        return [t.name for t in device.tags.all()]

    @check_perm("lava_scheduler_app.change_device")
    def delete(self, hostname, name):
        """
        Name
        ----
        `scheduler.devices.tags.delete` (`hostname`, `name`)

        Description
        -----------
        [superuser only]
        Remove a device tag from the device

        Arguments
        ---------
        `hostname`: string
          Device hostname
        `name`: string
          Tag name to remove
        Return value
        ------------
        None
        """
        try:
            device = Device.objects.get(hostname=hostname)
        except Device.DoesNotExist:
            raise xmlrpc.client.Fault(404, "Device '%s' was not found." % hostname)

        try:
            tag = Tag.objects.get(name=name)
        except Tag.DoesNotExist:
            raise xmlrpc.client.Fault(404, "Tag '%s' was not found." % name)

        device.tags.remove(tag)
        device.save()



# -*- coding: utf-8 -*-
# Generated by Django 1.9.4 on 2016-08-05 09:45
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("lava_results_app", "0012_namedtestattribute_metadata"),
    ]

    operations = [
        migrations.CreateModel(
            name="BugLink",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("url", models.URLField(max_length=1024, verbose_name="Bug Link URL")),
                ("object_id", models.PositiveIntegerField()),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.ContentType",
                    ),
                ),
            ],
        ),
        migrations.AlterUniqueTogether(
            name="buglink", unique_together=set([("object_id", "url", "content_type")])
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.4 on 2016-04-13 16:10
import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("lava_results_app", "0008_auto_20160405_1543")]

    operations = [
        migrations.AddField(
            model_name="query",
            name="limit",
            field=models.PositiveIntegerField(
                default=200,
                validators=[django.core.validators.MinValueValidator(20)],
                verbose_name="Results limit",
            ),
        )
    ]



# Copyright (C) 2014 Linaro Limited
#
# Author: Neil Williams <neil.williams@linaro.org>
#
# This file is part of LAVA Dispatcher.
#
# LAVA Dispatcher is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# LAVA Dispatcher is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along
# with this program; if not, see <http://www.gnu.org/licenses>.


import os
import yaml
import logging
import unittest
from lava_dispatcher.device import NewDevice
from lava_dispatcher.parser import JobParser
from lava_dispatcher.actions.boot.u_boot import UBootAction, UBootSecondaryMedia
from lava_dispatcher.actions.boot import BootloaderCommandOverlay
from lava_dispatcher.actions.deploy.apply_overlay import CompressRamdisk
from lava_dispatcher.actions.deploy.tftp import TftpAction
from lava_dispatcher.job import Job
from lava_dispatcher.action import Pipeline
from lava_common.exceptions import JobError
from lava_dispatcher.tests.test_basic import Factory, StdoutTestCase
from lava_dispatcher.tests.utils import DummyLogger, infrastructure_error
from lava_dispatcher.utils.network import dispatcher_ip
from lava_dispatcher.utils.filesystem import tftpd_dir
from lava_dispatcher.utils.strings import substitute


class UBootFactory(Factory):  # pylint: disable=too-few-public-methods
    """
    Not Model based, this is not a Django factory.
    Factory objects are dispatcher based classes, independent
    of any database objects.
    """

    def create_bbb_job(self, filename):
        return self.create_job("bbb-03.jinja2", filename)

    def create_x15_job(self, filename):
        return self.create_job("x15-01.jinja2", filename)

    def create_juno_job(self, filename):
        return self.create_job("juno-r2-01.jinja2", filename)

    def create_zcu102_job(self, filename):
        return self.create_job("zcu102.jinja2", filename)


class TestUbootAction(StdoutTestCase):  # pylint: disable=too-many-public-methods
    def setUp(self):
        super().setUp()
        self.factory = UBootFactory()

    @unittest.skipIf(infrastructure_error("mkimage"), "u-boot-tools not installed")
    def test_simulated_action(self):
        job = self.factory.create_bbb_job("sample_jobs/uboot-ramdisk.yaml")
        self.assertIsNotNone(job)

        # uboot and uboot-ramdisk have the same pipeline structure
        description_ref = self.pipeline_reference("uboot.yaml", job=job)
        self.assertEqual(description_ref, job.pipeline.describe(False))

        self.assertIsNone(job.validate())

    def test_tftp_pipeline(self):
        job = self.factory.create_bbb_job("sample_jobs/uboot-ramdisk.yaml")
        self.assertEqual(
            [action.name for action in job.pipeline.actions],
            ["tftp-deploy", "uboot-action", "lava-test-retry", "finalize"],
        )
        tftp = [
            action for action in job.pipeline.actions if action.name == "tftp-deploy"
        ][0]
        self.assertTrue(
            tftp.get_namespace_data(action=tftp.name, label="tftp", key="ramdisk")
        )
        self.assertIsNotNone(tftp.internal_pipeline)
        self.assertEqual(
            [action.name for action in tftp.internal_pipeline.actions],
            [
                "download-retry",
                "download-retry",
                "download-retry",
                "prepare-tftp-overlay",
                "lxc-create-udev-rule-action",
                "deploy-device-env",
            ],
        )
        self.assertIn(
            "ramdisk",
            [
                action.key
                for action in tftp.internal_pipeline.actions
                if hasattr(action, "key")
            ],
        )
        self.assertIn(
            "kernel",
            [
                action.key
                for action in tftp.internal_pipeline.actions
                if hasattr(action, "key")
            ],
        )
        self.assertIn(
            "dtb",
            [
                action.key
                for action in tftp.internal_pipeline.actions
                if hasattr(action, "key")
            ],
        )
        self.assertNotIn("=", tftpd_dir())
        job.validate()
        tftp.validate()
        self.assertEqual([], tftp.errors)

    def test_device_bbb(self):
        job = self.factory.create_bbb_job("sample_jobs/uboot.yaml")
        self.assertEqual(
            job.device["commands"]["connections"]["uart0"]["connect"],
            "telnet localhost 6000",
        )
        self.assertEqual(job.device["commands"].get("interrupt", " "), " ")
        methods = job.device["actions"]["boot"]["methods"]
        self.assertIn("u-boot", methods)
        self.assertEqual(
            methods["u-boot"]["parameters"].get("bootloader_prompt"), "U-Boot"
        )

    @unittest.skipIf(infrastructure_error("mkimage"), "u-boot-tools not installed")
    def test_uboot_action(self):
        job = self.factory.create_bbb_job("sample_jobs/uboot-ramdisk.yaml")
        job.validate()
        self.assertEqual(job.pipeline.errors, [])
        self.assertIn("u-boot", job.device["actions"]["boot"]["methods"])
        params = job.device["actions"]["deploy"]["parameters"]
        self.assertIn("mkimage_arch", params)
        boot_message = params.get(
            "boot_message", job.device.get_constant("kernel-start-message")
        )
        self.assertIsNotNone(boot_message)
        for action in job.pipeline.actions:
            action.validate()
            if isinstance(action, UBootAction):
                self.assertIn("method", action.parameters)
                self.assertEqual("u-boot", action.parameters["method"])
                self.assertEqual(
                    "reboot: Restarting system",
                    action.parameters.get("parameters", {}).get(
                        "shutdown-message", job.device.get_constant("shutdown-message")
                    ),
                )
            if isinstance(action, TftpAction):
                self.assertIn("ramdisk", action.parameters)
                self.assertIn("kernel", action.parameters)
                self.assertIn("to", action.parameters)
                self.assertEqual("tftp", action.parameters["to"])
            if isinstance(action, CompressRamdisk):
                self.assertEqual(action.mkimage_arch, "arm")
            self.assertTrue(action.valid)

    @unittest.skipIf(infrastructure_error("lxc-start"), "lxc-start not installed")
    def test_fastboot_uboot(self):  # pylint: disable=too-many-locals
        job = self.factory.create_x15_job("sample_jobs/x15-uboot.yaml")
        job.validate()
        description_ref = self.pipeline_reference("x15-uboot.yaml", job=job)
        self.assertEqual(description_ref, job.pipeline.describe(False))
        deploy = [
            action
            for action in job.pipeline.actions
            if action.name == "fastboot-deploy"
        ][0]
        enter = [
            action
            for action in deploy.internal_pipeline.actions
            if action.name == "uboot-enter-fastboot"
        ][0]
        interrupt = [
            action
            for action in enter.internal_pipeline.actions
            if action.name == "bootloader-interrupt"
        ][0]
        self.assertIsNotNone(interrupt.params)
        self.assertNotEqual(interrupt.params, {})
        self.assertEqual("u-boot", interrupt.method)

    @unittest.skipIf(infrastructure_error("lxc-start"), "lxc-start not installed")
    def test_x15_uboot_nfs(self):  # pylint: disable=too-many-locals
        job = self.factory.create_x15_job("sample_jobs/x15-nfs.yaml")
        job.validate()
        description_ref = self.pipeline_reference("x15-nfs.yaml", job=job)
        self.assertEqual(description_ref, job.pipeline.describe(False))
        tftp_deploy = [
            action for action in job.pipeline.actions if action.name == "tftp-deploy"
        ][0]
        prepare = [
            action
            for action in tftp_deploy.internal_pipeline.actions
            if action.name == "prepare-tftp-overlay"
        ][0]
        nfs = [
            action
            for action in prepare.internal_pipeline.actions
            if action.name == "extract-nfsrootfs"
        ][0]
        self.assertIn("compression", nfs.parameters["nfsrootfs"])
        self.assertEqual(nfs.parameters["nfsrootfs"]["compression"], "gz")

    def test_juno_uboot_nfs(self):
        job = self.factory.create_juno_job("sample_jobs/juno-uboot-nfs.yaml")
        job.validate()
        description_ref = self.pipeline_reference("juno-uboot-nfs.yaml", job=job)
        self.assertEqual(description_ref, job.pipeline.describe(False))

    def test_overlay_action(self):  # pylint: disable=too-many-locals
        parameters = {
            "device_type": "beaglebone-black",
            "job_name": "uboot-pipeline",
            "job_timeout": "15m",
            "action_timeout": "5m",
            "priority": "medium",
            "actions": {
                "boot": {
                    "method": "u-boot",
                    "commands": "ramdisk",
                    "type": "bootz",
                    "prompts": ["linaro-test", "root@debian:~#"],
                },
                "deploy": {
                    "ramdisk": "initrd.gz",
                    "kernel": "zImage",
                    "dtb": "broken.dtb",
                },
            },
        }
        device = NewDevice(
            os.path.join(os.path.dirname(__file__), "../devices/bbb-01.yaml")
        )
        job = Job(4212, parameters, None)
        job.device = device
        pipeline = Pipeline(job=job, parameters=parameters["actions"]["boot"])
        job.pipeline = pipeline
        overlay = BootloaderCommandOverlay()
        pipeline.add_action(overlay)
        ip_addr = dispatcher_ip(None)
        parsed = []
        kernel_addr = job.device["parameters"][overlay.parameters["type"]]["ramdisk"]
        ramdisk_addr = job.device["parameters"][overlay.parameters["type"]]["ramdisk"]
        dtb_addr = job.device["parameters"][overlay.parameters["type"]]["dtb"]
        kernel = parameters["actions"]["deploy"]["kernel"]
        ramdisk = parameters["actions"]["deploy"]["ramdisk"]
        dtb = parameters["actions"]["deploy"]["dtb"]

        substitution_dictionary = {
            "{SERVER_IP}": ip_addr,
            # the addresses need to be hexadecimal
            "{KERNEL_ADDR}": kernel_addr,
            "{DTB_ADDR}": dtb_addr,
            "{RAMDISK_ADDR}": ramdisk_addr,
            "{BOOTX}": "%s %s %s %s"
            % (overlay.parameters["type"], kernel_addr, ramdisk_addr, dtb_addr),
            "{RAMDISK}": ramdisk,
            "{KERNEL}": kernel,
            "{DTB}": dtb,
        }
        params = device["actions"]["boot"]["methods"]
        params["u-boot"]["ramdisk"]["commands"] = substitute(
            params["u-boot"]["ramdisk"]["commands"], substitution_dictionary
        )

        commands = params["u-boot"]["ramdisk"]["commands"]
        self.assertIs(type(commands), list)
        self.assertIn("setenv loadkernel 'tftp ${kernel_addr_r} zImage'", commands)
        self.assertIn(
            "setenv loadinitrd 'tftp ${initrd_addr_r} initrd.gz; setenv initrd_size ${filesize}'",
            commands,
        )
        self.assertIn("setenv loadfdt 'tftp ${fdt_addr_r} broken.dtb'", commands)
        self.assertNotIn("setenv kernel_addr_r '{KERNEL_ADDR}'", commands)
        self.assertNotIn("setenv initrd_addr_r '{RAMDISK_ADDR}'", commands)
        self.assertNotIn("setenv fdt_addr_r '{DTB_ADDR}'", commands)

        for line in params["u-boot"]["ramdisk"]["commands"]:
            line = line.replace("{SERVER_IP}", ip_addr)
            # the addresses need to be hexadecimal
            line = line.replace("{KERNEL_ADDR}", kernel_addr)
            line = line.replace("{DTB_ADDR}", dtb_addr)
            line = line.replace("{RAMDISK_ADDR}", ramdisk_addr)
            line = line.replace(
                "{BOOTX}",
                "%s %s %s %s"
                % (overlay.parameters["type"], kernel_addr, ramdisk_addr, dtb_addr),
            )
            line = line.replace("{RAMDISK}", ramdisk)
            line = line.replace("{KERNEL}", kernel)
            line = line.replace("{DTB}", dtb)
            parsed.append(line)
        self.assertIn("setenv loadkernel 'tftp ${kernel_addr_r} zImage'", parsed)
        self.assertIn(
            "setenv loadinitrd 'tftp ${initrd_addr_r} initrd.gz; setenv initrd_size ${filesize}'",
            parsed,
        )
        self.assertIn("setenv loadfdt 'tftp ${fdt_addr_r} broken.dtb'", parsed)
        self.assertNotIn("setenv kernel_addr_r '{KERNEL_ADDR}'", parsed)
        self.assertNotIn("setenv initrd_addr_r '{RAMDISK_ADDR}'", parsed)
        self.assertNotIn("setenv fdt_addr_r '{DTB_ADDR}'", parsed)

    def test_boot_commands(self):
        job = self.factory.create_bbb_job(
            "sample_jobs/uboot-ramdisk-inline-commands.yaml"
        )
        job.validate()
        uboot = [
            action for action in job.pipeline.actions if action.name == "uboot-action"
        ][0]
        overlay = [
            action
            for action in uboot.internal_pipeline.actions
            if action.name == "bootloader-overlay"
        ][0]
        self.assertEqual(
            overlay.commands,
            ["a list", "of commands", "with a {KERNEL_ADDR} substitution"],
        )

    @unittest.skipIf(infrastructure_error("xnbd-server"), "xnbd-server not installed")
    def test_nbd_boot(self):
        job = self.factory.create_bbb_job("sample_jobs/bbb-initrd-nbd.yaml")
        job.validate()
        self.assertEqual(job.pipeline.errors, [])
        description_ref = self.pipeline_reference("bbb-initrd-nbd.yaml", job=job)
        self.assertEqual(description_ref, job.pipeline.describe(False))
        # Fixme: more asserts
        self.assertIn("u-boot", job.device["actions"]["boot"]["methods"])
        params = job.device["actions"]["deploy"]["parameters"]
        self.assertIsNotNone(params)
        for action in job.pipeline.actions:
            action.validate()
            if isinstance(action, UBootAction):
                self.assertIn("method", action.parameters)
                self.assertEqual("u-boot", action.parameters["method"])
            elif isinstance(action, TftpAction):
                self.assertIn("initrd", action.parameters)
                self.assertIn("kernel", action.parameters)
                self.assertIn("nbdroot", action.parameters)
                self.assertIn("to", action.parameters)
                self.assertEqual("nbd", action.parameters["to"])
            self.assertTrue(action.valid)
        uboot = [
            action for action in job.pipeline.actions if action.name == "uboot-action"
        ][0]
        overlay = [
            action
            for action in uboot.internal_pipeline.actions
            if action.name == "bootloader-overlay"
        ][0]
        for setenv in overlay.commands:
            if "setenv nbdbasekargs" in setenv:
                self.assertIn("rw", setenv.split("'")[1])
                self.assertIn("${extraargs}", setenv.split("'")[1])
                self.assertEqual(3, len(setenv.split("'")))

    def test_transfer_media(self):
        """
        Test adding the overlay to existing rootfs
        """
        job = self.factory.create_bbb_job(
            "sample_jobs/uboot-ramdisk-inline-commands.yaml"
        )
        job.validate()
        description_ref = self.pipeline_reference(
            "uboot-ramdisk-inline-commands.yaml", job=job
        )
        self.assertEqual(description_ref, job.pipeline.describe(False))
        uboot = [
            action for action in job.pipeline.actions if action.name == "uboot-action"
        ][0]
        retry = [
            action
            for action in uboot.internal_pipeline.actions
            if action.name == "uboot-retry"
        ][0]
        transfer = [
            action
            for action in retry.internal_pipeline.actions
            if action.name == "overlay-unpack"
        ][0]
        self.assertIn("transfer_overlay", transfer.parameters)
        self.assertIn("download_command", transfer.parameters["transfer_overlay"])
        self.assertIn("unpack_command", transfer.parameters["transfer_overlay"])

    def test_download_action(self):
        job = self.factory.create_bbb_job("sample_jobs/uboot.yaml")
        for action in job.pipeline.actions:
            action.validate()
            self.assertTrue(action.valid)
        job.validate()
        self.assertEqual(job.pipeline.errors, [])
        deploy = None
        overlay = None
        extract = None
        for action in job.pipeline.actions:
            if action.name == "tftp-deploy":
                deploy = action
        if deploy:
            for action in deploy.internal_pipeline.actions:
                if action.name == "prepare-tftp-overlay":
                    overlay = action
        if overlay:
            for action in overlay.internal_pipeline.actions:
                if action.name == "extract-nfsrootfs":
                    extract = action
        test_dir = overlay.get_namespace_data(
            action="test", label="results", key="lava_test_results_dir"
        )
        self.assertIsNotNone(test_dir)
        self.assertIn("/lava-", test_dir)
        self.assertIsNotNone(extract)
        self.assertEqual(extract.timeout.duration, 240)

    def test_reset_actions(self):
        job = self.factory.create_bbb_job("sample_jobs/uboot.yaml")
        uboot_action = None
        uboot_retry = None
        reset_action = None
        for action in job.pipeline.actions:
            action.validate()
            self.assertTrue(action.valid)
            if action.name == "uboot-action":
                uboot_action = action
        names = [r_action.name for r_action in uboot_action.internal_pipeline.actions]
        self.assertIn("connect-device", names)
        self.assertIn("uboot-retry", names)
        for action in uboot_action.internal_pipeline.actions:
            if action.name == "uboot-retry":
                uboot_retry = action
        names = [r_action.name for r_action in uboot_retry.internal_pipeline.actions]
        self.assertIn("reset-device", names)
        self.assertIn("bootloader-interrupt", names)
        self.assertIn("expect-shell-connection", names)
        self.assertIn("bootloader-commands", names)
        for action in uboot_retry.internal_pipeline.actions:
            if action.name == "reset-device":
                reset_action = action
        names = [r_action.name for r_action in reset_action.internal_pipeline.actions]
        self.assertIn("pdu-reboot", names)

    def test_secondary_media(self):
        """
        Test UBootSecondaryMedia validation
        """
        job_parser = JobParser()
        (rendered, _) = self.factory.create_device("cubie1.jinja2")
        cubie = NewDevice(yaml.safe_load(rendered))
        sample_job_file = os.path.join(
            os.path.dirname(__file__), "sample_jobs/cubietruck-removable.yaml"
        )
        sample_job_data = open(sample_job_file)
        job = job_parser.parse(sample_job_data, cubie, 4212, None, "")
        job.logger = DummyLogger()
        job.validate()
        sample_job_data.close()
        uboot_action = [
            action
            for action in job.pipeline.actions
            if action.name == "uboot-action"
            and action.parameters["namespace"] == "boot2"
        ][0]
        u_boot_media = [
            action
            for action in uboot_action.internal_pipeline.actions
            if action.name == "uboot-from-media"
            and action.parameters["namespace"] == "boot2"
        ][0]
        self.assertIsInstance(u_boot_media, UBootSecondaryMedia)
        self.assertEqual([], u_boot_media.errors)
        self.assertEqual(
            u_boot_media.parameters["kernel"], "/boot/vmlinuz-3.16.0-4-armmp-lpae"
        )
        self.assertEqual(
            u_boot_media.parameters["kernel"],
            u_boot_media.get_namespace_data(
                action="download-action", label="file", key="kernel"
            ),
        )
        self.assertEqual(
            u_boot_media.parameters["ramdisk"],
            u_boot_media.get_namespace_data(
                action="compress-ramdisk", label="file", key="ramdisk"
            ),
        )
        self.assertEqual(
            u_boot_media.parameters["dtb"],
            u_boot_media.get_namespace_data(
                action="download-action", label="file", key="dtb"
            ),
        )
        # use the base class name so that uboot-from-media can pick up the value reliably.
        self.assertEqual(
            u_boot_media.parameters["root_uuid"],
            u_boot_media.get_namespace_data(
                action="bootloader-from-media", label="uuid", key="root"
            ),
        )
        device = u_boot_media.get_namespace_data(
            action="storage-deploy", label="u-boot", key="device"
        )
        self.assertIsNotNone(device)
        part_reference = "%s:%s" % (
            job.device["parameters"]["media"]["usb"][device]["device_id"],
            u_boot_media.parameters["boot_part"],
        )
        self.assertEqual(
            part_reference,
            u_boot_media.get_namespace_data(
                action=u_boot_media.name, label="uuid", key="boot_part"
            ),
        )
        self.assertEqual(part_reference, "0:1")

    def test_xz_nfs(self):
        job = self.factory.create_bbb_job("sample_jobs/uboot-nfs.yaml")
        # this job won't validate as the .xz nfsrootfs URL is a fiction
        self.assertRaises(JobError, job.validate)
        tftp_deploy = [
            action for action in job.pipeline.actions if action.name == "tftp-deploy"
        ][0]
        prepare = [
            action
            for action in tftp_deploy.internal_pipeline.actions
            if action.name == "prepare-tftp-overlay"
        ][0]
        nfs = [
            action
            for action in prepare.internal_pipeline.actions
            if action.name == "extract-nfsrootfs"
        ][0]
        self.assertIn("compression", nfs.parameters["nfsrootfs"])
        self.assertEqual(nfs.parameters["nfsrootfs"]["compression"], "xz")

    def test_prefix(self):
        job = self.factory.create_bbb_job("sample_jobs/bbb-skip-install.yaml")
        job.validate()
        tftp_deploy = [
            action for action in job.pipeline.actions if action.name == "tftp-deploy"
        ][0]
        prepare = [
            action
            for action in tftp_deploy.internal_pipeline.actions
            if action.name == "prepare-tftp-overlay"
        ][0]
        nfs = [
            action
            for action in prepare.internal_pipeline.actions
            if action.name == "extract-nfsrootfs"
        ][0]
        self.assertIn("prefix", nfs.parameters["nfsrootfs"])
        self.assertEqual(nfs.parameters["nfsrootfs"]["prefix"], "jessie/")
        self.assertEqual(nfs.param_key, "nfsrootfs")

    def test_zcu102(self):
        job = self.factory.create_zcu102_job("sample_jobs/zcu102-ramdisk.yaml")
        job.validate()
        self.assertEqual(job.pipeline.errors, [])
        description_ref = self.pipeline_reference("zcu102-ramdisk.yaml", job=job)
        self.assertEqual(description_ref, job.pipeline.describe(False))

    @unittest.skipIf(infrastructure_error("lxc-start"), "lxc-start not installed")
    def test_imx8m(self):
        job = self.factory.create_job("imx8m-01.jinja2", "sample_jobs/imx8m.yaml")
        self.assertIsNotNone(job)
        job.validate()
        self.assertEqual(job.pipeline.errors, [])
        description_ref = self.pipeline_reference("imx8m.yaml", job=job)
        self.assertEqual(description_ref, job.pipeline.describe(False))
        deploy = [
            action
            for action in job.pipeline.actions
            if action.name == "fastboot-deploy"
        ][0]
        fastboot = [
            action
            for action in deploy.internal_pipeline.actions
            if action.name == "uboot-enter-fastboot"
        ][0]
        bootloader = [
            action
            for action in fastboot.internal_pipeline.actions
            if action.name == "bootloader-interrupt"
        ][0]
        self.assertEqual("u-boot", bootloader.method)


class TestKernelConversion(StdoutTestCase):
    def setUp(self):
        logger = logging.getLogger("dispatcher")
        logger.disabled = True
        logger.propagate = False
        logger = logging.getLogger("lava-dispatcher")
        logger.disabled = True
        logger.propagate = False
        self.device = NewDevice(
            os.path.join(os.path.dirname(__file__), "../devices/bbb-01.yaml")
        )
        bbb_yaml = os.path.join(
            os.path.dirname(__file__), "sample_jobs/uboot-ramdisk.yaml"
        )
        with open(bbb_yaml) as sample_job_data:
            self.base_data = yaml.safe_load(sample_job_data)
        self.deploy_block = [
            block for block in self.base_data["actions"] if "deploy" in block
        ][0]["deploy"]
        self.boot_block = [
            block for block in self.base_data["actions"] if "boot" in block
        ][0]["boot"]
        self.parser = JobParser()

    def test_zimage_bootz(self):
        self.deploy_block["kernel"]["type"] = "zimage"
        job = self.parser.parse(yaml.dump(self.base_data), self.device, 4212, None, "")
        job.logger = DummyLogger()
        job.validate()
        deploy = [
            action for action in job.pipeline.actions if action.name == "tftp-deploy"
        ][0]
        overlay = [
            action
            for action in deploy.internal_pipeline.actions
            if action.name == "prepare-tftp-overlay"
        ][0]
        prepare = [
            action
            for action in overlay.internal_pipeline.actions
            if action.name == "prepare-kernel"
        ][0]
        uboot_prepare = [
            action
            for action in prepare.internal_pipeline.actions
            if action.name == "uboot-prepare-kernel"
        ][0]
        self.assertEqual("zimage", uboot_prepare.kernel_type)
        self.assertEqual("bootz", uboot_prepare.bootcommand)
        self.assertFalse(uboot_prepare.mkimage_conversion)

    def test_image(self):
        self.deploy_block["kernel"]["type"] = "image"
        job = self.parser.parse(yaml.dump(self.base_data), self.device, 4212, None, "")
        job.logger = DummyLogger()
        job.validate()
        deploy = [
            action for action in job.pipeline.actions if action.name == "tftp-deploy"
        ][0]
        overlay = [
            action
            for action in deploy.internal_pipeline.actions
            if action.name == "prepare-tftp-overlay"
        ][0]
        prepare = [
            action
            for action in overlay.internal_pipeline.actions
            if action.name == "prepare-kernel"
        ][0]
        uboot_prepare = [
            action
            for action in prepare.internal_pipeline.actions
            if action.name == "uboot-prepare-kernel"
        ][0]
        self.assertEqual("image", uboot_prepare.kernel_type)
        # bbb-01.yaml does not contain booti parameters, try to convert to a uImage
        self.assertEqual("bootm", uboot_prepare.bootcommand)
        self.assertTrue(uboot_prepare.mkimage_conversion)

    def test_uimage(self):
        self.deploy_block["kernel"]["type"] = "uimage"
        job = self.parser.parse(yaml.dump(self.base_data), self.device, 4212, None, "")
        job.logger = DummyLogger()
        job.validate()
        deploy = [
            action for action in job.pipeline.actions if action.name == "tftp-deploy"
        ][0]
        overlay = [
            action
            for action in deploy.internal_pipeline.actions
            if action.name == "prepare-tftp-overlay"
        ][0]
        prepare = [
            action
            for action in overlay.internal_pipeline.actions
            if action.name == "prepare-kernel"
        ][0]
        uboot_prepare = [
            action
            for action in prepare.internal_pipeline.actions
            if action.name == "uboot-prepare-kernel"
        ][0]
        self.assertEqual("uimage", uboot_prepare.kernel_type)
        self.assertEqual("bootm", uboot_prepare.bootcommand)
        self.assertFalse(uboot_prepare.mkimage_conversion)

    def test_zimage_nobootz(self):
        # drop bootz from the device for this part of the test
        del self.device["parameters"]["bootz"]
        self.deploy_block["kernel"]["type"] = "zimage"
        job = self.parser.parse(yaml.dump(self.base_data), self.device, 4212, None, "")
        job.logger = DummyLogger()
        job.validate()
        deploy = [
            action for action in job.pipeline.actions if action.name == "tftp-deploy"
        ][0]
        overlay = [
            action
            for action in deploy.internal_pipeline.actions
            if action.name == "prepare-tftp-overlay"
        ][0]
        prepare = [
            action
            for action in overlay.internal_pipeline.actions
            if action.name == "prepare-kernel"
        ][0]
        uboot_prepare = [
            action
            for action in prepare.internal_pipeline.actions
            if action.name == "uboot-prepare-kernel"
        ][0]
        self.assertEqual("zimage", uboot_prepare.kernel_type)
        self.assertEqual("bootm", uboot_prepare.bootcommand)
        self.assertTrue(uboot_prepare.mkimage_conversion)

    def test_uimage_boot_type(self):
        # uimage in boot type
        del self.deploy_block["kernel"]["type"]
        self.boot_block["type"] = "bootm"
        job = self.parser.parse(yaml.dump(self.base_data), self.device, 4212, None, "")
        job.logger = DummyLogger()
        job.validate()
        deploy = [
            action for action in job.pipeline.actions if action.name == "tftp-deploy"
        ][0]
        overlay = [
            action
            for action in deploy.internal_pipeline.actions
            if action.name == "prepare-tftp-overlay"
        ][0]
        self.assertNotIn(
            "uboot-prepare-kernel",
            [action.name for action in overlay.internal_pipeline.actions],
        )


class TestOverlayCommands(StdoutTestCase):  # pylint: disable=too-many-public-methods
    def setUp(self):
        super().setUp()
        self.factory = UBootFactory()

    def test_combined_ramdisk_nfs(self):
        job = self.factory.create_bbb_job("sample_jobs/bbb-ramdisk-nfs.yaml")
        tftp_deploy = [
            action for action in job.pipeline.actions if action.name == "tftp-deploy"
        ][0]
        prepare = [
            action
            for action in tftp_deploy.internal_pipeline.actions
            if action.name == "prepare-tftp-overlay"
        ][0]
        nfs = [
            action
            for action in prepare.internal_pipeline.actions
            if action.name == "extract-nfsrootfs"
        ][0]
        ramdisk = [
            action
            for action in prepare.internal_pipeline.actions
            if action.name == "extract-overlay-ramdisk"
        ][0]
        modules = [
            action
            for action in prepare.internal_pipeline.actions
            if action.name == "extract-modules"
        ][0]
        overlay = [
            action
            for action in prepare.internal_pipeline.actions
            if action.name == "apply-overlay-tftp"
        ][0]
        self.assertIsNotNone(ramdisk.parameters.get("ramdisk"))
        self.assertIsNotNone(ramdisk.parameters["ramdisk"].get("url"))
        self.assertIsNotNone(ramdisk.parameters["ramdisk"].get("compression"))
        self.assertTrue(ramdisk.parameters["ramdisk"].get("install_modules", True))
        self.assertTrue(ramdisk.parameters["ramdisk"].get("install_overlay", True))
        self.assertIsNotNone(modules.parameters.get("ramdisk"))
        self.assertIsNotNone(modules.parameters.get("nfsrootfs"))
        self.assertIsNotNone(nfs.parameters.get("nfsrootfs"))
        self.assertIsNotNone(overlay.parameters.get("nfsrootfs"))
        self.assertIsNotNone(overlay.parameters.get("ramdisk"))

    def test_ramdisk_nfs_nomodules(self):
        job = self.factory.create_bbb_job("sample_jobs/bbb-uinitrd-nfs.yaml")
        tftp_deploy = [
            action for action in job.pipeline.actions if action.name == "tftp-deploy"
        ][0]
        prepare = [
            action
            for action in tftp_deploy.internal_pipeline.actions
            if action.name == "prepare-tftp-overlay"
        ][0]
        nfs = [
            action
            for action in prepare.internal_pipeline.actions
            if action.name == "extract-nfsrootfs"
        ][0]
        ramdisk = [
            action
            for action in prepare.internal_pipeline.actions
            if action.name == "extract-overlay-ramdisk"
        ][0]
        modules = [
            action
            for action in prepare.internal_pipeline.actions
            if action.name == "extract-modules"
        ][0]
        overlay = [
            action
            for action in prepare.internal_pipeline.actions
            if action.name == "apply-overlay-tftp"
        ][0]
        self.assertIsNotNone(ramdisk.parameters.get("ramdisk"))
        self.assertIsNotNone(ramdisk.parameters["ramdisk"].get("url"))
        self.assertIsNone(ramdisk.parameters["ramdisk"].get("compression"))
        self.assertFalse(ramdisk.parameters["ramdisk"].get("install_overlay", True))
        self.assertIsNotNone(modules.parameters.get("ramdisk"))
        self.assertIsNotNone(modules.parameters.get("nfsrootfs"))
        self.assertIsNotNone(nfs.parameters.get("nfsrootfs"))
        self.assertIsNotNone(overlay.parameters.get("nfsrootfs"))
        self.assertIsNotNone(overlay.parameters.get("ramdisk"))



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2017-12-05 16:26
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("lava_scheduler_app", "0032_add_state_health_to_testjob_and_device")
    ]

    operations = [
        migrations.RemoveField(model_name="devicestatetransition", name="created_by"),
        migrations.RemoveField(model_name="devicestatetransition", name="device"),
        migrations.RemoveField(model_name="devicestatetransition", name="job"),
        migrations.RemoveField(model_name="device", name="current_job"),
        migrations.RemoveField(model_name="device", name="health_status"),
        migrations.RemoveField(model_name="device", name="status"),
        migrations.RemoveField(model_name="notification", name="job_status_trigger"),
        migrations.RemoveField(model_name="testjob", name="requested_device"),
        migrations.RemoveField(model_name="testjob", name="status"),
        migrations.RemoveField(model_name="testjob", name="submit_token"),
        migrations.DeleteModel(name="DeviceStateTransition"),
    ]



# -*- coding: utf-8 -*-
# Copyright (C) 2017-2018 Linaro Limited
#
# Author: Neil Williams <neil.williams@linaro.org>
#
# This file is part of LAVA.
#
# LAVA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation
#
# LAVA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with LAVA.  If not, see <http://www.gnu.org/licenses/>.

from django.contrib.sites.models import Site
from django.core.management.base import BaseCommand, CommandError, CommandParser


class Command(BaseCommand):
    help = "Update Django Site"

    def add_arguments(self, parser):
        cmd = self

        class SubParser(CommandParser):
            """
            Sub-parsers constructor that mimic Django constructor.
            See http://stackoverflow.com/a/37414551
            """

            def __init__(self, **kwargs):
                super().__init__(cmd, **kwargs)

        sub = parser.add_subparsers(
            dest="sub_command", help="Sub commands", parser_class=SubParser
        )
        sub.required = True

        list_parser = sub.add_parser("list", help="List the current Site")

        update_parser = sub.add_parser("update", help="Update site properties")
        update_parser.add_argument("--name", type=str, help="Display name of the Site")
        update_parser.add_argument(
            "--domain", type=str, default=None, help="site domain"
        )

    def handle(self, *args, **options):
        if options["sub_command"] == "list":
            self.handle_list()
        elif options["sub_command"] == "update":
            self.handle_update(options["name"], options["domain"])

    def handle_list(self):
        """ List the current Site """
        site = Site.objects.get_current()
        self.stdout.write("Site:")
        self.stdout.write("\tDomain: %s" % site.domain)
        self.stdout.write("\tName: %s" % site.name)

    def handle_update(self, name, domain):
        """ Update Site properties """
        site = Site.objects.get_current()
        site.domain = domain
        site.name = name
        site.save()



# -*- coding: utf-8 -*-
# Copyright (C) 2011-2018 Linaro Limited
#
# Author: Neil Williams <neil.williams@linaro.org>
#         Remi Duraffort <remi.duraffort@linaro.org>
#
# This file is part of LAVA.
#
# LAVA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation
#
# LAVA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with LAVA.  If not, see <http://www.gnu.org/licenses/>.

from django.conf.urls import url
from lava_scheduler_app.views import (
    active_device_list,
    active_jobs,
    device_derestrict_device,
    device_detail,
    device_edit_description,
    device_health,
    device_type_health_history_log,
    mydevices_health_history_log,
    device_list,
    device_reports,
    device_restrict_device,
    all_device_types,
    device_type_detail,
    device_dictionary,
    device_dictionary_plain,
    maintenance_devices,
    device_type_reports,
    edit_worker_desc,
    failure_report,
    favorite_jobs,
    job_fetch_data,
    health_job_list,
    healthcheck,
    index,
    job_annotate_failure,
    job_cancel,
    job_fail,
    job_change_priority,
    job_definition,
    job_definition_plain,
    job_description_yaml,
    job_detail,
    job_list,
    job_configuration,
    job_errors,
    job_log_file_plain,
    job_log_incremental,
    job_timing,
    job_resubmit,
    job_status,
    job_submit,
    job_toggle_favorite,
    lab_health,
    longest_jobs,
    multinode_job_definition,
    multinode_job_definition_plain,
    mydevice_list,
    myjobs,
    online_device_list,
    passing_health_checks,
    queue,
    reports,
    running,
    username_list_json,
    worker_detail,
    worker_health,
    workers,
    download_device_type_template,
    similar_jobs,
)


urlpatterns = [  # pylint: disable=invalid-name
    url(r"^$", index, name="lava.scheduler"),
    url(r"^reports$", reports, name="lava.scheduler.reports"),
    url(r"^reports/failures$", failure_report, name="lava.scheduler.failure_report"),
    url(r"^edit_worker_desc", edit_worker_desc, name="lava.scheduler.edit_worker_desc"),
    url(r"^activejobs$", active_jobs, name="lava.scheduler.job.active"),
    url(r"^alljobs$", job_list, name="lava.scheduler.job.list"),
    url(r"joberrors$", job_errors, name="lava.scheduler.job.errors"),
    url(r"^jobsubmit$", job_submit, name="lava.scheduler.job.submit"),
    url(r"^device_types$", all_device_types, name="lava.scheduler.device_types"),
    url(
        r"^device_type/(?P<pk>[-_a-zA-Z0-9]+)$",
        device_type_detail,
        name="lava.scheduler.device_type.detail",
    ),
    url(r"^alldevices$", device_list, name="lava.scheduler.alldevices"),
    url(
        r"^device/(?P<pk>[-_a-zA-Z0-9.@]+)$",
        device_detail,
        name="lava.scheduler.device.detail",
    ),
    url(
        r"^device/(?P<pk>[-_a-zA-Z0-9.@]+)/devicedict$",
        device_dictionary,
        name="lava.scheduler.device.dictionary",
    ),
    url(
        r"^device/(?P<pk>[-_a-zA-Z0-9.@]+)/devicedict/plain$",
        device_dictionary_plain,
        name="lava.scheduler.device.dictionary.plain",
    ),
    url(r"^allworkers$", workers, name="lava.scheduler.workers"),
    url(
        r"^worker/(?P<pk>[-_a-zA-Z0-9.@]+)$",
        worker_detail,
        name="lava.scheduler.worker.detail",
    ),
    url(
        r"^worker/(?P<pk>[-_a-zA-Z0-9.@]+)/health$",
        worker_health,
        name="lava.scheduler.worker.health",
    ),
    url(r"^labhealth/$", lab_health, name="lava.scheduler.labhealth"),
    url(
        r"^labhealth/device/(?P<pk>[-_a-zA-Z0-9.@]+)$",
        health_job_list,
        name="lava.scheduler.labhealth.detail",
    ),
    url(r"^longestjobs$", longest_jobs, name="lava.scheduler.longest_jobs"),
    url(
        r"^job/(?P<pk>[0-9]+|[0-9]+\.[0-9]+)$",
        job_detail,
        name="lava.scheduler.job.detail",
    ),
    url(
        r"^job/(?P<pk>[0-9]+|[0-9]+\.[0-9]+)/definition$",
        job_definition,
        name="lava.scheduler.job.definition",
    ),
    url(
        r"^job/(?P<pk>[0-9]+|[0-9]+\.[0-9]+)/definition/plain$",
        job_definition_plain,
        name="lava.scheduler.job.definition.plain",
    ),
    url(
        r"^job/(?P<pk>[0-9]+|[0-9]+\.[0-9]+)/description$",
        job_description_yaml,
        name="lava.scheduler.job.description.yaml",
    ),
    url(
        r"^job/(?P<pk>[0-9]+|[0-9]+\.[0-9]+)/multinode_definition$",
        multinode_job_definition,
        name="lava.scheduler.job.multinode_definition",
    ),
    url(
        r"^job/(?P<pk>[0-9]+|[0-9]+\.[0-9]+)/multinode_definition/plain$",
        multinode_job_definition_plain,
        name="lava.scheduler.job.multinode_definition.plain",
    ),
    url(
        r"^job/(?P<pk>[0-9]+|[0-9]+\.[0-9]+)/configuration$",
        job_configuration,
        name="lava.scheduler.job.configuration",
    ),
    url(
        r"^job/(?P<pk>[0-9]+|[0-9]+\.[0-9]+)/log_file/plain$",
        job_log_file_plain,
        name="lava.scheduler.job.log_file.plain",
    ),
    url(
        r"^job/(?P<pk>[0-9]+|[0-9]+\.[0-9]+)/timing$",
        job_timing,
        name="lava.scheduler.job.timing",
    ),
    url(
        r"^job/(?P<pk>[0-9]+|[0-9]+\.[0-9]+)/job_status$",
        job_status,
        name="lava.scheduler.job_status",
    ),
    url(
        r"^job/(?P<pk>[0-9]+|[0-9]+\.[0-9]+)/cancel$",
        job_cancel,
        name="lava.scheduler.job.cancel",
    ),
    url(
        r"^job/(?P<pk>[0-9]+|[0-9]+\.[0-9]+)/fail$",
        job_fail,
        name="lava.scheduler.job.fail",
    ),
    url(
        r"^job/(?P<pk>[0-9]+|[0-9]+\.[0-9]+)/resubmit$",
        job_resubmit,
        name="lava.scheduler.job.resubmit",
    ),
    url(
        r"^job/(?P<pk>[0-9]+|[0-9]+\.[0-9]+)/annotate_failure$",
        job_annotate_failure,
        name="lava.scheduler.job.annotate_failure",
    ),
    url(
        r"^job/(?P<pk>[0-9]+|[0-9]+\.[0-9]+)/toggle_favorite$",
        job_toggle_favorite,
        name="lava.scheduler.job.toggle_favorite",
    ),
    url(
        r"^job/(?P<pk>[0-9]+|[0-9]+\.[0-9]+)/log_pipeline_incremental$",
        job_log_incremental,
        name="lava.scheduler.job.log_incremental",
    ),
    url(
        r"^job/(?P<pk>[0-9]+|[0-9]+\.[0-9]+)/job_data$",
        job_fetch_data,
        name="lava.scheduler.job.fetch_data",
    ),
    url(r"^myjobs$", myjobs, name="lava.scheduler.myjobs"),
    url(r"^favorite-jobs$", favorite_jobs, name="lava.scheduler.favorite_jobs"),
    url(
        r"^job/(?P<pk>[0-9]+|[0-9]+\.[0-9]+)/priority$",
        job_change_priority,
        name="lava.scheduler.job.priority",
    ),
    url(
        r"^device/(?P<pk>[-_a-zA-Z0-9.@]+)/restrict$",
        device_restrict_device,
        name="lava.scheduler.device.restrict_device",
    ),
    url(
        r"^device/(?P<pk>[-_a-zA-Z0-9.@]+)/derestrict$",
        device_derestrict_device,
        name="lava.scheduler.device.derestrict_device",
    ),
    url(
        r"^device/(?P<pk>[-_a-zA-Z0-9.@]+)/description$",
        device_edit_description,
        name="lava.scheduler.device.edit_description",
    ),
    url(
        r"^device/(?P<pk>[-_a-zA-Z0-9.@]+)/health$",
        device_health,
        name="lava.scheduler.device.health",
    ),
    url(
        r"^alldevices/active$", active_device_list, name="lava.scheduler.active_devices"
    ),
    url(
        r"^alldevices/online$", online_device_list, name="lava.scheduler.online_devices"
    ),
    url(
        r"^alldevices/passinghealthchecks$",
        passing_health_checks,
        name="lava.scheduler.passing_health_checks",
    ),
    url(
        r"^alldevices/maintenance$",
        maintenance_devices,
        name="lava.scheduler.maintenance_devices",
    ),
    url(
        r"^reports/device/(?P<pk>[-_a-zA-Z0-9.@]+)",
        device_reports,
        name="lava.scheduler.device_report",
    ),
    url(
        r"^reports/device_type/(?P<pk>[-_a-zA-Z0-9]+)",
        device_type_reports,
        name="lava.scheduler.device_type_report",
    ),
    url(r"^mydevices$", mydevice_list, name="lava.scheduler.mydevice_list"),
    url(
        r"^username-list-json$",
        username_list_json,
        name="lava.scheduler.username_list_json",
    ),
    url(r"^queue$", queue, name="lava.scheduler.queue"),
    url(r"^healthcheck$", healthcheck, name="lava.scheduler.healthcheck"),
    url(r"^running$", running, name="lava.scheduler.running"),
    url(
        r"^dthealthhistory/device_type/(?P<pk>[-_a-zA-Z0-9]+)",
        device_type_health_history_log,
        name="lava.scheduler.device_type_health_history_log",
    ),
    url(
        r"^mydevicetypehealthhistory$",
        mydevices_health_history_log,
        name="lava.scheduler.mydevices_health_history_log",
    ),
    url(
        r"^devicetypeyaml/(?P<pk>[-_a-zA-Z0-9]+)",
        download_device_type_template,
        name="lava_scheduler_download_device_type_yaml",
    ),
    url(
        r"^job/(?P<pk>[0-9]+|[0-9]+.[0-9]+)/similarjobs$",
        similar_jobs,
        name="lava.scheduler.job.similar_jobs",
    ),
]
